#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np

class GaussianRing(Omega):
    r'''Gaussian ring polymer intra-molecular correlation function
    
    
    **Mathematical Definition**

    .. math::
    
        \hat{\omega}(k) = 1+2N^{-1}\sum_{t=1}^{N-1}(N-t)\exp(\frac{-k^2\sigma^2t(N-t)}{6N})
         

    **Variable Definitions**

        - :math:`\hat{\omega}(k)` 
            *intra*-molecular correlation function at wavenumber :math:`k`

        - :math:`N`
            number of monomers/sites in gaussian ring
 
        - :math:`\sigma`
            contact distance between sites (i.e. site diameter)


    **Description**
        
    The Gaussian ring is an ideal model for a cyclic chain
    that assumes a random walk between successive monomer
    segments along the chain, constrained such that ends join
    together to form a ring with no intra-molecular excluded
    volume.


    References
    ----------
    Schweizer, K.S.; Curro, J.G.; Integral-Equation Theory of Polymer Melts -
    Intramolecular Structure, Local Order, and the Correlation Hole,
    Macromolecules, 1988, 21 (10), pp 3070, doi:10.1021/ma00188a027

    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #calculate Fourier space domain and omega values
        domain = pyPRISM.domain(dr=0.1,length=1000)
        omega  = pyPRISM.omega.GaussianRing(sigma=1.0,length=100)
        x = domain.k
        y = omega.calculate(x)

        #plot it!
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()

        #Define a PRISM system and set omega(k) for type A
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.omega['A','A']  = pyPRISM.omega.GaussianRing(sigma=1.0,length=100)

    
    '''
    def __init__(self,sigma,length):
        r'''Constructor
        
        Arguments
        ---------
        sigma: float
            contact distance between sites (site diameter)        
    
        length: float
            number of monomers/sites in gaussian ring
        '''
        self.sigma = sigma
        self.length = length
        self.value = None
        
    def __repr__(self):
        return '<Omega: GaussianRing>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        self.value = np.zeros_like(k)
        ss = self.sigma * self.sigma
        kk = k*k
        for i in range(self.length):
            # for j in range(self.length):
            j = 0
            self.value += np.exp(-ss*kk*abs(i-j)*(self.length-abs(i-j))/(6.0*self.length))
        return self.value
        
        



#!python
from __future__ import division,print_function
import numpy as np
class AtomicClosure:
    '''Baseclass for all atomic closures
    
    Note
    ----
    Currently, this class doesn't do anything besides group all of the
    *atomic* closures under a single inheritance heirarchy. 
    '''
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.FreelyJointedChain import FreelyJointedChain
import unittest
import numpy as np

class FreelyJointedChain_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a FreelyJointedChain Omega?'''
        k = np.arange(0.75,3.5,0.05)
        N = 100
        l = 4.0/3.0
        
        E = np.sin(k*l)/(k*l)
        N = N
        O1 = (1 - E*E - 2*E/N + (2*E**(N+1))/N)/((1-E)**2.0)
        
        G = FreelyJointedChain(N,l)
        O2 = G.calculate(k)
        np.testing.assert_array_almost_equal(O1,O2)
        
        



import warnings
import pint

class UnitConverter(object):
    r'''Unit conversion utility 
    
    **Description**

        pyPRISM operates in a system of reduced units commonly called 'Lennard
        Jones units'. In this unit system, all measures are reported relative
        to characteristic values: length = :math:`d_c`, mass = :math:`m_c`, and
        energy = :math:`e_c`. This class is designed to make some common
        conversions between reduced and real units easier.
        
        See the Theory.General notebook of the :ref:`tutorial` for a more detailed
        discussion of these units and how to work with the UnitConverter utility.

    .. note::

        The methods prefixed with "to" all expect floating point values in
        reduced units and convert to real units. 

    .. warning::

        This class uses the `Pint <https://pint.readthedocs.io>`_ package. While the
        rest of pyPRISM will function without Pint, this class will not be
        available without this dependency installed in the current environment. 

    
    Example
    -------
    .. code-block:: python

        import pyPRISM

        domain = pyPRISM.Domain(length=4096,dr=0.25)

        # create unit converter utility 
        uc = pyPRISM.util.UnitConverter(dc=1.5,dc_unit='nm')

        # convert wavenumber from LJ units to real units 
        # using built-in conversion
        real_k = uc.toInvAngstrom(domain.k)
        real_k_magnitudes = real_k.magnitude

        # convert radius in real units to reduced units
        # manually using pint
        real_radius = 123.5 # angstrom
        reduced_radius = real_radius * uc('angstrom').to('dc').magnitude


    
    '''
    def __init__(self,
                 dc=1.0,   dc_unit='nanometer',
                 mc=14.02, mc_unit='gram/mole',
                 ec=2.48,  ec_unit='kilojoule/mole'):
        r''' Constructor
        
        Arguments
        ---------
        dc,dc_unit: float,str
            Magnitude and unit of characteristic distance

        mc,mc_unit: float,str
            Magnitude and unit of characteristic mass

        ec,ec_unit: float,str
            Magnitude and unit of characteristic energy


        Attributes
        ----------
        d,dc: Pint Quantity
            The defined characteristic distance as a defined Pint Quantity.

        m,mc: Pint Quantity
            The defined characteristic mass as a defined Pint Quantity.

        e,ec: Pint Quantity
            The defined characteristic energy as a defined Pint Quantity.

        '''

        self.pint = pint.UnitRegistry()

        dString = 'dchar = {} {} = dc'.format(dc,dc_unit)
        mString = 'mchar = {} {} = mc'.format(mc,mc_unit)
        eString = 'echar = {} {} = ec'.format(ec,ec_unit)
        # print('--> Defining reduced unit {}'.format(dString))
        # print('--> Defining reduced unit {}'.format(mString))
        # print('--> Defining reduced unit {}'.format(eString))
        self.pint.define(dString)
        self.pint.define(mString)
        self.pint.define(eString)

        self.d = self.dc = self.pint.Quantity(1,'dc')
        self.m = self.mc = self.pint.Quantity(1,'mc')
        self.e = self.ec = self.pint.Quantity(1,'ec')

    def __repr__(self):
        return '<UnitConverter  dc:{} | mc:{} | ec:{}>'.format(self.dc.to_base_units(),self.mc.to_base_units(),self.ec.to_base_units())
    
    def __call__(self,unit_string):
        '''Convenience method for accessing the pint UnitRegistry'''
        return self.pint(unit_string)

    def toKelvin(self,temperature):
        r'''Convert thermal energy to temperature units in :math:`K`

        Arguments
        ---------
        temperature: float or np.ndarray of floats
            Value of thermal energy (:math:`k_{B}T`) to be converted

        Returns
        -------
        temperature: Pint Quantity
            temperature in :math:`K` as a Pint Quantity. Use the magnitude
            attribute (temperature.magnitude) to obtain the numerical value of
            the temperature as a floating point value.
        '''
        new_value = (temperature*self.e)/self.pint('boltzmann_constant')

        # need to handle both the molar and non-molar characteristic energy
        try:
            return new_value.to('K')
        except pint.errors.DimensionalityError:
            new_value /= self.pint('N_A')
            return new_value.to('K')
    def toCelcius(self,temperature):
        r'''Convert thermal energy to temperature units in Celcius

        Arguments
        ---------
        temperature: float or np.ndarray of floats
            Value of thermal energy (:math:`k_{B}T`) to be converted

        Returns
        -------
        temperature: Pint Quantity
            temperature in Kelvin as a Pint Quantity. Use the magnitude
            attribute (temperature.magnitude) to obtain the numerical value of
            the temperature as a floating point value.
        '''
        return self.toKelvin(temperature).to('C')
    def toInvAngstrom(self,wavenumber):
        r'''Convert wavenumbers to real units

        Arguments
        ---------
        wavenumber: float, or np.ndarray of floats
            Value(s) of wavenumbers to be converted

        Returns
        -------
        wavenumber: Pint Quantity
            wavenumbers in :math:`AA^{-1}` as a Pint Quantity. Use the
            magnitude attribute (wavenumber.magnitude) to obtain the numerical
            value of the wavenumber as a floating point value.
        '''
        new_value = wavenumber*(1.0/self.d)
        return new_value.to('angstrom^-1')
    def toInvNanometer(self,wavenumber):
        r'''Convert wavenumbers to real units

        Arguments
        ---------
        wavenumber: float, or np.ndarray of floats
            Value(s) of wavenumbers to be converted

        Returns
        -------
        wavenumber: Pint Quantity
            wavenumbers in :math:`nm^{-1}` as a Pint Quantity. Use the
            magnitude attribute (wavenumber.magnitude) to obtain the numerical
            value of the wavenumber as a floating point value.
        '''
        return self.toInvAngstrom(wavenumber).to('nanometer^-1')
    def toConcentration(self,density):
        r'''Convert reduced number density to real concentration units

        Arguments
        ---------
        density: float, or np.ndarray of floats
            Value(s) of density to be converted

        Returns
        -------
        concentration: Pint Quantity
            density in :math:`mol/L` as a Pint Quantity. Use the
            magnitude attribute (concentration.magnitude) to obtain the
            numerical value of the contration as a floating point value.
        '''
        new_value = density/(self.d**3.0)/self.pint('N_A')
        return new_value.to('mol/L')
    def toVolumeFraction(self,density,diameter):
        r'''Convert reduced number density to volume fraction

        Arguments
        ---------
        density: float, or np.ndarray of floats
            Value(s) of density to be converted

        diameter: float
            diameter of the site associated with the density

        Returns
        -------
        vol_fraction: Pint Quantity
            unitless volume as a Pint Quantity. Use the magnitude attribute
            (vol_fraction.magnitude) to obtain the numerical value of the
            volume fraction as a floating point value.
        '''
        new_value = density*4.0/3.0 * self.pint('pi') * (diameter/2.0)**(3.0)
        return new_value*self.pint('unitless')






#!python
from __future__ import division,print_function
from pyPRISM.core.Space import Space
import numpy as np
from scipy.fftpack import dst

class Domain(object):
    r'''Define domain and transform between Real and Fourier space

    **Mathematical Definition**
        
        The continuous, 1-D, radially symmetric Fourier transform is written
        as follows:

        .. math::

            k\ \hat{f}(k) = 4 \pi \int r\ f(r) \sin(k\ r) dr

        We define the following discretizations

        .. math:: 

            r = (i+1)\Delta r

            k = (j+1)\Delta k

            \Delta k = \frac{\pi}{\Delta r (N + 1)}

        to yield

        .. math::

            \hat{F}_j = 4 \pi \Delta r \sum_{i=0}^{N-1} F_i \sin\left(\frac{\pi}{N+1} (i+1)(j+1)\right)

        with the following definitions:

        .. math::

            \hat{F}_j = (j+1)\ \Delta k\ \hat{f}((j+1)\Delta k) = k \hat{f}(k)

        .. math::

            F_i = (i+1)\Delta r\ f((i+1)\Delta r) = r f(r)

        The above equations describe a Real to Real, type-II discrete sine
        transform (DST). To tranform to and from Fourier space we will use the
        type-II and type-III DST's respectively. With Scipy's interface to
        fftpack, the following functional coeffcients are

        .. math::

            C^{DSTII} = 2 \pi r \Delta r

        .. math::

            C^{DSTIII} = \frac{k \Delta k}{4 \pi^2}
    
    
    **Description**

        Domain describes the discretization of Real and Fourier space
        and also sets up the functions and coefficients for transforming
        data between them.
    
    '''
    def __init__(self,length,dr=None,dk=None):
        r'''Constructor

        Arguments
        ---------
        length: int
            Number of gridpoints in Real and Fourier space grid

        dr,dk: float
            Grid spacing in Real space or Fourier space. Only one can be
            specified as it fixes the other.

        '''
        self._length = length
        
        if (dr is None) and (dk is None):
            raise ValueError('Real or Fourier grid spacing must be specified')
            
        elif (dr is not None) and (dk is not None):
            raise ValueError('Cannot specify **both** Real and Fourier grid spacings independently.')
            
        elif dr is not None:
            self.dr = dr #dk is set in property setter
            
        elif dk is not None:
            self.dk = dk #dr is set in property setter
            
            
        self.build_grid() #build grid should have been called already but we'll be safe
    
    def build_grid(self):
        '''Construct the Real and Fourier Space grids and transform coefficients'''
        self.r = np.arange(self._dr,self._dr*(self._length+1),self._dr)
        self.k = np.arange(self.dk,self.dk*(self._length+1),self.dk)
        self.DST_II_coeffs = 2.0*np.pi *self.r*self._dr 
        self.DST_III_coeffs = self.k * self.dk/(4.0*np.pi*np.pi)
        self.long_r = self.r.reshape((-1,1,1))
    
    @property
    def dr(self):
        '''Real grid spacing'''
        return self._dr
    @dr.setter
    def dr(self,value):
        self._dr = value
        self._dk = np.pi/(self._dr*self._length)
        self.build_grid()#need to re-build grid since spacing has changed
    
    @property
    def dk(self):
        '''Fourier grid spacing'''
        return self._dk
    @dk.setter
    def dk(self,value):
        self._dk = value
        self._dr = np.pi/(self._dk*self._length)
        self.build_grid()#need to re-build grid since spacing has changed
        
    @property
    def length(self):
        '''Number of points in grid'''
        return self._length
    @length.setter
    def length(self,value):
        self._length = value
        self.build_grid()#need to re-build grid since length has changed
        
    def __repr__(self):
        return '<Domain length:{} dr/rmax:{:4.3f}/{:3.1f} dk/kmax:{:4.3f}/{:3.1f}>'.format(self.length,self.dr,self.r[-1],self.dk,self.k[-1])
    
    def to_fourier(self,array):
        r''' Discrete Sine Transform of a numpy array 
        
        Arguments
        ---------
        array: float ndarray
            Real-space data to be transformed
            
        Returns
        -------
        array: float ndarray
            data transformed to fourier space


        Peforms a Real-to-Real Discrete Sine Transform  of type II 
        on a numpy array of non-complex values. For radial data that is 
        symmetric in :math:`\phi` and :math`\theta`, this is a correct transform
        to go from Real-space to Fourier-space. 
        
        
        '''
        return dst(self.DST_II_coeffs*array,type=2)/self.k
    
    def to_real(self,array):
        ''' Discrete Sine Transform of a numpy array 
        
        Arguments
        ---------
        array: float ndarray
            Fourier-space data to be transformed
            
        Returns
        -------
        array: float ndarray
            data transformed to Real space

        Peforms a Real-to-Real Discrete Sine Transform  of type III
        on a numpy array of non-complex values. For radial data that is 
        symmetric in :math:`\phi` and :math`\theta`, this is a correct transform
        to go from Real-space to Fourier-space. 
        
        '''
        return dst(self.DST_III_coeffs*array,type=3)/self.r
    
    def MatrixArray_to_fourier(self,marray):
        ''' Transform all curves of a MatrixArray to Fourier space in-place

        Arguments
        ---------
        marray: :class:`pyPRISM.core.MatrixArray.MatrixArray`
            MatrixArray to be transformed

        Raises
        ------
        *ValueError*:
            If the supplied MatrixArray is already in Real-space
        '''
        if marray.space == Space.Fourier:
            raise ValueError('MatrixArray is marked as already in Fourier space')
            
        for (i,j),(t1,t2),curve in marray.itercurve():
            marray[t1,t2] = self.to_fourier(curve)
        
        marray.space = Space.Fourier
            
    def MatrixArray_to_real(self,marray):
        ''' Transform all curves of a MatrixArray to Real space in-place 

        Arguments
        ---------
        marray: :class:`pyPRISM.core.MatrixArray.MatrixArray`
            MatrixArray to be transformed

        Raises
        ------
        ValueError:
            If the supplied MatrixArray is already in Real-space
        '''
        if marray.space == Space.Real:
            raise ValueError('MatrixArray is marked as already in Real space')
            
        for (i,j),(t1,t2),curve in marray.itercurve():
            marray[t1,t2] = self.to_real(curve)
            
        marray.space = Space.Real
            



#!python
from __future__ import division,print_function
from pyPRISM.potential.LennardJones import LennardJones
import numpy as np
class WeeksChandlerAndersen(LennardJones):
    r'''Purely repulsive Weeks-Chandler-Andersen potential
    

    **Mathematical Definition**
    
    .. math::
   
	U_{\alpha,\beta}(r)  =
            \begin{cases}
                4\epsilon_{\alpha,\beta}\left[\left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{12.0}
                - \left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{6.0}\right] +
                \epsilon_{\alpha,\beta} & r<r_{cut}

                0.0      &  r \geq r_{cut} 
            \end{cases}

    .. math::
    
	r_{cut} = 2^{1/6}\sigma_{\alpha,\beta}    

    
    **Variable Definitions**
    
    :math:`\epsilon_{\alpha,\beta}`
        Strength of repulsion between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`\sigma_{\alpha,\beta}`
        Length scale of interaction between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`r`
        Distance between sites. 
    
    :math:`r_{cut}`
        Cutoff distance where the value of the potential goes to zero. 
    
   
    **Description**

    	The Weeks-Chandler-Andersen potential for purely repulsive interactions.
    	This potential is equivalent to the Lennard-Jones potential cut and
    	shifted at the minimum of the potential, which occurs at 
    	:math:`r=2^{1/6}\sigma`. 
    
    Example
    -------
    .. code-block:: python

        import pyPRISM
	
        #Define a PRISM system and set the A-B interaction potential
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.potential['A','B'] = pyPRISM.potential.WeeksChandlerAndersen(epsilon=1.0,sigma=1.0)

    
    .. warning::

        If sigma is specified such that it does not fall on the solution grid
        of the :class:`~pyPRISM.core.Domain.Domain` object specified in
        :class:`~pyPRISM.core.System.System`, then the sigma will effectively
        be rounded. A warning should be emitted during the construction of a
        :class:`~pyPRISM.core.PRISM.PRISM` object if this occurs.
    
    '''
    def __init__(self,epsilon,sigma=None):
        r''' Constructor
        
        Arguments
        ---------
    	epsilon: float
            Repulsive strength modifier
        
        sigma: float, *optional*
            Contact distance. If not specified, sigma will be calculated from 
            the diameters specified in the :class:`~pyPRISM.core.System.System`
            object.
    
        '''
        super(WeeksChandlerAndersen,self).__init__(epsilon=epsilon,sigma=sigma,rcut=True,shift=True)
        
    def __repr__(self):
        return '<Potential: WeeksChandlerAndersen>'

    def calculate(self,r):
        r'''Calculate value of potential

        Attributes
        ----------
        r: float np.ndarray
            Array of pair distances at which to calculate potential values
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'
        self.rcut = self.sigma * 2**(1.0/6.0)
        return super(WeeksChandlerAndersen,self).calculate(r)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.core.Space import Space
from pyPRISM.core.PairTable import PairTable
import numpy as np

def second_virial(PRISM,extrapolate=True):
    r'''Calculate the second virial coefficient

    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.

    extrapolate: bool, *optional*
        If *True*, extrapolate :math:`h_{\alpha,\beta}` to :math:`k=0` rather
        than reporting the value at the lowest-k. Defaults to *True*.
    
    Returns
    -------
    B2: pyPRISM.core.PairTable
        Pairtable of B2 values


    **Mathematical Definition**

    .. math::

        B_{2}^{\alpha,\beta} = -0.5 \hat{h}_{\alpha,\beta}(k=0)

    
    **Variable Definitions**

        - :math:`B_{2}^{\alpha,\beta}`
            Second virial coefficient between site types :math:`\alpha` and
            :math:`\beta`

        - :math:`\hat{h}_{\alpha,\beta}(k)`
            Fourier-space total correlation function between site types
            :math:`\alpha` and :math:`\beta` at wavevector :math:`k`


    **Description**

        The second virial coefficient (:math:`B_{2}^{\alpha,\beta}`) is a 
        thermodynamic descriptor related to the pairwise interactions between 
        components in a system. In general, :math:`B_{2}^{\alpha,\beta}>0`
        signifies repulsive interactions between site types :math:`\alpha` and
        :math:`\beta`, and :math:`B_{2}^{\alpha,\beta}<0` signifies attractive
        interactions. For example, in a polymer-solvent system, one definition
        of the theta condition is when :math:`B_{2}^{\alpha,\beta}=0`.


    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        B2 = pyPRISM.calculate.second_virial(PRISM)

        B2_BB = B2['B','B']
    
    '''
    
    if PRISM.totalCorr.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.totalCorr)
    
    B2 = PairTable(name='B2',types=PRISM.sys.types)
    for i,t1 in enumerate(PRISM.sys.types):
        for j,t2 in enumerate(PRISM.sys.types):
            if extrapolate:
                x = PRISM.sys.domain.k[:3]
                y = - 0.5 * PRISM.totalCorr[t1,t2][:3]
    
                fit = np.poly1d(np.polyfit(x,y,2))

                B2[t1,t2] = fit(0)
            else:
                B2[t1,t2] = - 0.5 * PRISM.totalCorr[t1,t2][0]

    return B2



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np


class FreelyJointedChain(Omega):
    r'''Freely jointed chain intra-molecular correlation function

    **Mathematical Definition**

    .. math::
    
        \hat{\omega}(k) = \frac{1 - E^2 - \frac{2E}{N} + \frac{2E^{N+1}}{N}}{(1-E)^2}

    .. math::
         
         E = \frac{\sin(k l)}{k l}


    **Variable Definitions**

        - :math:`\hat{\omega}(k)` 
            *intra*-molecular correlation function at wavenumber :math:`k`

        - :math:`N`
            number of repeat units in chain
 
        - :math:`l`
            bond-length


    **Description**
        
        The freely-jointed chain is an ideal polymer chain model
        that assumes a constant bond length :math:`l` and no correlations
        between the directions of different bond vectors 
        (i.e. :math:`<cos(\theta_{ij})>=0`). In other words, monomer segments
        are assumed to have no intra-molecular excluded volume.


    References
    ----------
    #. Schweizer, K.S.; Curro, J.G.; Integral-Equation Theory of Polymer Melts
       - Intramolecular Structure, Local Order, and the Correlation Hole,
       Macromolecules, 1988, 21 (10), pp 3070
       [`link <https://doi.org/10.1021/ma00188a027>`__]

    #. Rubinstein, M; Colby, R.H; Polymer Physics. 2003. Oxford University Press.

    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #calculate Fourier space domain and omega values
        domain = pyPRISM.domain(dr=0.1,length=1000)
        omega  = pyPRISM.omega.FreelyJointedChain(length=100,l=1.0)
        x = domain.k
        y = omega.calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()
	
	#define a PRISM system and set omega(k) for type A
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.omega['A','A']  = pyPRISM.omega.FreelyJointedChain(length=100,l=1.0)

    
    '''
    def __init__(self,length,l):
        r'''Constructor
        
        Arguments
        ---------
        length: float
            number of monomers/sites in Freely-jointed chain
            
        l: float
            bond length
        '''
        self.length = self.N = length
        self.l = l
        self.value = None
        
    def __repr__(self):
        return '<Omega: FreelyJointedChain>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        E = np.sin(k*self.l)/(k*self.l)
        N = self.N
        self.value = (1 - E*E - 2*E/N + (2*E**(N+1))/N)/((1-E)**2.0)
        return self.value
        
        
class FJC(FreelyJointedChain):
    '''Alias of FreelyJointedChain'''
    pass



#!python
from __future__ import division,print_function
import unittest
from pyPRISM.potential.Exponential import Exponential
import numpy as np

class Exponential_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a Exponential potential?'''
        r = np.arange(0.75,3.5,0.05)
        sigma   = 1.05
        epsilon = 0.25
        alpha = 0.25
        
        U1  =  -epsilon * np.exp(-(r-sigma)/(alpha))
        U1 = np.where(r>sigma,U1,1e6)
        
        U2 = Exponential(sigma=sigma,epsilon=epsilon,alpha=alpha).calculate(r)
        np.testing.assert_array_almost_equal(U1,U2)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
from pyPRISM.omega.FreelyJointedChain import FreelyJointedChain
import numpy as np
import scipy.integrate 

import warnings

NFJC_WARNING = '''
The numerical integrations required for the NFJC omega
calculation are slow and scale poorly with chain length 
so this omega may take minutes or longer to calculate 
even for modest chain lengths e.g. N=200.'''

class NonOverlappingFreelyJointedChain(Omega):
    r'''Freely jointed chain with excluded volume intra-molecular correlation function


    .. warning:: 
        The numerical integrations required for the NFJC omega
        calculation are slow and scale poorly with chain length 
        so this omega may take minutes or longer to calculate 
        even for modest chain lengths e.g. N=200.

    
    **Mathematical Definition**

    .. math::
    
        \hat{\omega}(k) = \hat{\omega}_{id}(k)+\frac{2}{N}\sum_{\tau=2}^{N-1}(N-\tau)
			[\hat{\omega}_{\tau}(k)-(\sin(k)/k)^{\tau}]
    
    .. math::
    
        \tau = |\alpha-\beta| 


    **Variable Definitions**

        - :math:`\hat{\omega}(k)` 
            *intra*-molecular correlation function at wavenumber :math:`k`

        - :math:`\hat{\omega}_{id}(k)` 
            *intra*-molecular correlation function for the ideal freely-jointed 
	    chain at wavenumber :math:`k`. Please see equation (15)
	    of Reference [1] for the mathematical representation.
        
        - :math:`\hat{\omega}_{\tau}(k)` 
            Please see equations (17,18,21) of Reference [1] for the
            mathematical representation.
	
	- :math:`N`
            number of repeat units in chain
 
        - :math:`\tau`
            number of monomers along chain separating sites :math:`\alpha` 
	    and :math:`\beta`. 


    **Description**
        
        The non-overlapping freely-jointed chain is an adjustment to the ideal
        freely jointed chain model that includes the effects of the excluded
        volume of monomer segments (i.e. bonds are not free to rotate over all
        angles). This model assumes a constant bond length :math:`l`. 


    References
    ----------
    #. Schweizer, K.S.; Curro, J.G.; Integral-Equation Theory of Polymer Melts
       - Intramolecular Structure, Local Order, and the Correlation Hole,
       Macromolecules, 1988, 21 (10), pp 3070
       [`link <https://doi.org/10.1021/ma00188a027>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #calculate Fourier space domain and omega values
        domain = pyPRISM.domain(dr=0.1,length=1000)
        omega  = pyPRISM.omega.NonOverlappingFreelyJointedChain(length=100,l=1.0)
        x = domain.k
        y = omega.calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()
	
	#define a PRISM system and set omega(k) for type A
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.omega['A','A']  = pyPRISM.omega.NonOverlappingFreelyJointedChain(length=100,l=1.0)

    
    '''
    def __init__(self,length,l):
        r'''Constructor
        
        Arguments
        ---------
        length: float
            number of monomers/sites in non-overlapping freely-jointed chain
            
        l: float
            bond length
        '''
        self.length = self.N = length
        self.l = l
        self.FJC = FreelyJointedChain(length=length,l=l)
        self.value = None

        warnings.warn(NFJC_WARNING)
        
    def __repr__(self):
        return '<Omega: NonOverlappingFreelyJointedChain>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        integrate = np.trapz
        integrate = scipy.integrate.simps
        self.value = np.zeros_like(k)

        B = []
        dx = 0.1
        x = np.arange(dx,100,dx)
        K,X = np.meshgrid(k,x,indexing='ij')
        # Precaculate factors for efficiency 
        ZBase = (1/(np.pi*K))*X*(np.sin(K-X)/(K-X) - np.sin(K+X)/(K+X))
        sinxx = np.sin(x)/x
        sinXX = np.sin(X)/X
        sinkk = np.sin(k)/k
        J0Base = (np.sin(x)/x - np.cos(x))
        for tau in range(2,self.length):
            J0val = 2/np.pi * integrate((sinxx)**(tau)*J0Base,x=x)
            B = (1 - J0val)**(-1.0)
            Z = ZBase * (sinXX)**(tau) 
            Jvals = integrate(Z,x=x,axis=1)
            omega_t = B * ((sinkk)**(tau) - Jvals)
            self.value +=  (self.length - tau) * (omega_t - (sinkk)**(tau))

        self.value  *= 2.0/self.length 
        self.value  += self.FJC.calculate(k)


        return self.value

class NFJC(NonOverlappingFreelyJointedChain):
    ''' Alias of NonOverlappingFreelyJointedChain '''
    pass
        
        



#!python
'''
PRISM uses pairwise decomposed potentials to describe the interactions between
site-types. pyPRISM provides the potentials listed below.  If you have
potential that is not listed, please consider filing an `Issue
<https://github.com/usnistgov/pyPRISM/issues>`_ on GitHub, implementing the
potential, and sharing with the community. See :ref:`contribute`
for more details.  
'''
from pyPRISM.potential.Potential import Potential
from pyPRISM.potential.Exponential import Exponential
from pyPRISM.potential.HardSphere import HardSphere
from pyPRISM.potential.LennardJones import LennardJones
from pyPRISM.potential.HardCoreLennardJones import HardCoreLennardJones
from pyPRISM.potential.WeeksChandlerAndersen import WeeksChandlerAndersen




#!python
from __future__ import division,print_function
from pyPRISM.closure.AtomicClosure import AtomicClosure
import numpy as np
class PercusYevick(AtomicClosure):
    r'''Percus Yevick closure evaluated in terms of a change of variables

    **Mathematial Definition**

        .. math:: c_{\alpha,\beta}(r) = (\exp(-U_{\alpha,\beta}(r)) - 1.0) (1.0 + \gamma_{\alpha,\beta}(r))

        .. math:: \gamma_{\alpha,\beta}(r) =  h_{\alpha,\beta}(r) - c_{\alpha,\beta}(r)

    **Variables Definitions**

        - :math:`h_{\alpha,\beta}(r)` 
            Total correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`c_{\alpha,\beta}(r)`
            Direct correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`U_{\alpha,\beta}(r)`
            Interaction potential value at distance :math:`r` between sites
            :math:`\alpha` and :math:`\beta`.
    

    **Description**

        The Percus-Yevick (PY) is derived by expanding the exponential of the
        direct correlation function, :math:`c_{\alpha,\beta}(r)`, in powers of 
        density shift from a reference state. See Reference [1] for a full
        derivation.
        
        The change of variables is necessary in order to use potentials with
        hard cores in the computational setting. Written in the standard form,
        this closure diverges with divergent potentials, which makes it
        impossible to numerically solve. 

        This closure has been shown to be accurate for systems with hard cores
        (strongly repulsive at short distances) and when the potential is short
        ranged. 
    
    References
    ----------
    #. Hansen, J.P.; McDonald, I.R.; Theory of Simple Liquids; Chapter 4, Section 4; 
       4th Edition (2013), Elsevier [`link
       <https://www.sciencedirect.com/science/book/9780123870322>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        sys.closure['A','A'] = pyPRISM.closure.PercusYevick()
        sys.closure['A','B'] = pyPRISM.closure.PercusYevick()
        sys.closure['B','B'] = pyPRISM.closure.HypernettedChain()

        # ** finish populating system object **

        PRISM = sys.createPRISM()

        PRISM.solve()
    
    '''
    def __init__(self,apply_hard_core=False):
        '''Contstructor

        Parameters
        ----------
        apply_hard_core: bool
            If True, the total correlation function will be assumed to be -1
            inside the core (:math:`r_{i,j}<(d_i + d_j)/2.0`) and the closure
            will not be applied in this region.
        '''
        self.potential = None
        self.value = None
        self.sigma = None
        self.apply_hard_core=apply_hard_core
        
    def __repr__(self):
        return '<AtomicClosure: PercusYevick>'
    
    def calculate(self,r,gamma):
        '''Calculate direct correlation function based on supplied :math:`\gamma`

        Arguments
        ---------
        r: np.ndarray
            array of real-space values associated with :math:`\gamma`

        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'
        
        if self.apply_hard_core:
            assert self.sigma is not None, 'If apply_hard_core=True, sigma parameter must be set!'

            # apply hard core condition 
            self.value = -1 - gamma

            # calculate closure outside hard core
            mask = r>self.sigma
            self.value[mask] = (np.exp(-self.potential[mask])-1.0)*(1.0+gamma[mask])
        else:
            self.value = (np.exp(-self.potential)-1.0)*(1.0+gamma)

        
        return self.value
        
        
class PY(PercusYevick):
    '''Alias of PercusYevick'''
    pass



#!python
r'''
The fundamental PRISM equation is written in Fourier space as

.. math::
    \hat{H}(k) = \hat{\Omega}(k)\hat{C}(k) \left[\hat{\Omega}(k) + \hat{H}(k)\right]

where :math:`\hat{H}(k)` is the total correlation function,
:math:`\hat{\Omega}(k)` is the *intra*-molecular correlation function, and
:math:`\hat{C}(k)` is the direct correlation function. At each wavenumber
:math:`k`, each of these variables is an :math:`n \times n` matrix of values,
where :math:`n` is the number of components or site types in the system.  The
goal of any PRISM calculation is to obtain the full set of partial correlation
functions. Using these correlation functions, a number of structural and
thermodynamic properties can be calculated.

The :py:mod:`pyPRISM.core` module holds the fundamental data structures that
carry out the PRISM calculation.

The :py:mod:`pyPRISM.calculate` module provides a number of functions which use
*solved* :class:`pyPRISM.core.PRISM` objects to calculate structural and
thermodynamic parameters.

The :py:mod:`pyPRISM.closure` module provides closure objects which are necessary for
solving the PRISM equations.

The :py:mod:`pyPRISM.omega` module provides analytical *intra*-molecular
correlation (:math:`\hat{\omega}(k)`) functions along with methods for loading
them from memory or files.

The :py:mod:`pyPRISM.potential` module provides pair potentials for describing
the *inter*-molecular interactions in a system. Pairwise interactions are also
how the chemistry of the system is described.

The :py:mod:`pyPRISM.trajectory` module contains classes for working with
molecular simulation trajectories.

The :py:mod:`pyPRISM.util` module provides various global helper functions
which do not fall under the above categories.

See the :ref:`tutorial` for more information on the details of using pyPRISM
and the PRISM theory formalism.
'''
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.IdentityMatrixArray import IdentityMatrixArray
from pyPRISM.core.Space import Space
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.ValueTable import ValueTable
from pyPRISM.core.System import System
from pyPRISM.core.Domain import Domain
from pyPRISM.core.Density import Density
from pyPRISM.core.Diameter import Diameter

from pyPRISM import calculate
from pyPRISM import closure
from pyPRISM import potential
from pyPRISM import omega

from pyPRISM import util

from pyPRISM.version import *

def test():
    r"""
    Run all tests using pytest.
    """
    import os
    import pytest
    path = os.path.split(__file__)[0]
    pytest.main(['-x',path])




#!python
'''
In PRISM, the molecular structure of molecules is encoded into
*intra*-molecular correlation functions called :math:`\hat{\omega}(k)`. All
connectivity and *intra*-molecular excluded volume is contained in these
functions. During a PRISM calculation, :math:`\hat{\omega}(k)` are specified as
input and are held fixed during the solution procedure. While this leads to a
decoupling of the *intra*-molecular and *inter*-molecular correlations within a
given PRISM calculation, methods such as self-consistent PRISM offer a route to
mitigating this potential problem. See :ref:`scprism` for more details. 

pyPRISM offers several analytical form factors which are listed below. If you
have an analytical :math:`\hat{\omega}(k)` that is not listed, please consider
filing an `Issue <https://github.com/usnistgov/pyPRISM/issues>`_ on GitHub,
implementing the :math:`\hat{\omega}(k)` and sharing with the community. See
:ref:`contribute` for more details.  

Alternatively, if no analytical form exists, :math:`\hat{\omega}(k)` can be
calculated using a simulation. The :class:`pyPRISM.trajectory.Debyer` class
implements the Debye summation method for calculating :math:`\hat{\omega}(k)`
from simulation. 

Finally, the :class:`~pyPRISM.omega.FromArray` and
:class:`~pyPRISM.omega.FromFile` classes exist for loading
:math:`\hat{\omega}(k)` calculated in memory or from another program.
'''
from pyPRISM.omega.Omega import Omega

from pyPRISM.omega.SingleSite import SingleSite
from pyPRISM.omega.NoIntra import NoIntra
from pyPRISM.omega.InterMolecular import InterMolecular

from pyPRISM.omega.FromFile import FromFile
from pyPRISM.omega.FromArray import FromArray

from pyPRISM.omega.Gaussian import Gaussian
from pyPRISM.omega.GaussianRing import GaussianRing

from pyPRISM.omega.DiscreteKoyama import DiscreteKoyama

from pyPRISM.omega.FreelyJointedChain import FreelyJointedChain
from pyPRISM.omega.FreelyJointedChain import FJC

from pyPRISM.omega.NonOverlappingFreelyJointedChain import NonOverlappingFreelyJointedChain
from pyPRISM.omega.NonOverlappingFreelyJointedChain import NFJC




import unittest
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
import numpy as np

class PairTable_TestCase(unittest.TestCase):
    def test_get_set(self):
        '''Can we set and get values from the table?'''
        PT = PairTable(['A','B','C'],'density')
        PT['A','B'] = 0.4
        PT['A',['A','C']] = 0.25
        
        self.assertEqual(PT['A','B'],0.4)
        self.assertEqual(PT['B','A'],0.4)
        self.assertEqual(PT['A','A'],0.25)
        self.assertEqual(PT['A','C'],0.25)
        self.assertEqual(PT['C','A'],0.25)
        
    def test_check(self):
        '''Can we check to make sure the table is filled?'''
        PT = PairTable(['A','B','C'],'density')
        PT['A','B'] = 0.4
        PT['A',['A','C']] = 0.25
        self.assertRaises(ValueError,PT.check)
        
    def test_apply(self):
        '''Can we apply a function to the table?'''
        PT = PairTable(['A','B','C'],'density')
        PT['A','B'] = 0.4
        PT.setUnset(0.25)
        
        PT.apply(np.square)
        
        self.assertEqual(PT['A','B'],0.4*0.4)
        self.assertEqual(PT['B','A'],0.4*0.4)
        self.assertEqual(PT['A','A'],0.25*0.25)
        self.assertEqual(PT['A','C'],0.25*0.25)
        self.assertEqual(PT['C','A'],0.25*0.25)
        self.assertEqual(PT['C','C'],0.25*0.25)
        
        
        
        
    def test_setUnset(self):
        '''Can wet set all of the unset table values?'''
        PT = PairTable(['A','B','C'],'density')
        PT['A','B'] = 0.4
        PT.setUnset(0.25)
        
        self.assertEqual(PT['A','B'],0.4)
        self.assertEqual(PT['B','A'],0.4)
        self.assertEqual(PT['A','A'],0.25)
        self.assertEqual(PT['A','C'],0.25)
        self.assertEqual(PT['C','A'],0.25)
        self.assertEqual(PT['C','C'],0.25)
        
    def test_iter_full(self):
        '''Can we iterate over all of the table pairs?'''
        types = ['A','B','C']
        ntypes = len(types)
        PT = PairTable(types,'density')
        
        numericPairs = [
                         (0,0),(0,1),(0,2),
                         (1,0),(1,1),(1,2),
                         (2,0),(2,1),(2,2)
                       ]
        
        counter = 0
        for i,t,v in PT.iterpairs(full=True):
            if i in numericPairs:
                numericPairs.remove(i)
            counter+=1
            
        #did we visit all expected pairs?
        self.assertEqual(len(numericPairs),0)
        
        #sanity check, did we visit the correct number of pairs?
        self.assertEqual(counter,ntypes*ntypes)
        
    def test_iter_diagonal(self):
        '''Can we iterate over the upper triangle + diagonal?'''
        types = ['A','B','C']
        ntypes = len(types)
        PT = PairTable(types,'density')
        
        numericPairs = [
                         (0,0),(0,1),(0,2),
                               (1,1),(1,2),
                                     (2,2)
                       ]
        
        counter = 0
        for i,t,v in PT.iterpairs(diagonal=True):
            if i in numericPairs:
                numericPairs.remove(i)
            counter+=1
            
        #did we visit all expected pairs?
        self.assertEqual(len(numericPairs),0)
        
        #sanity check, did we visit the correct number of pairs?
        self.assertEqual(counter,ntypes*(ntypes+1)//2)
        
    def test_iter_triangle(self):
        '''Can we iterate over only the upper triangle?'''
        types = ['A','B','C']
        ntypes = len(types)
        PT = PairTable(types,'density')
        
        numericPairs = [
                         (0,1),(0,2),
                               (1,2),
                                    
                       ]
        
        counter = 0
        for i,t,v in PT.iterpairs(diagonal=False):
            if i in numericPairs:
                numericPairs.remove(i)
            counter+=1
            
        #did we visit all expected pairs?
        self.assertEqual(len(numericPairs),0)
        
        #sanity check, did we visit the correct number of pairs?
        self.assertEqual(counter,ntypes*(ntypes-1)//2)
    
    def test_MatrixArray_export(self):
        types = ['A','B','C']
        length = 1024
        rank = len(types)
        values1 = np.ones(length)
        values2 = np.ones(length)*5.0
        values3 = np.ones(length)*2.1234
        
        MA1 = MatrixArray(length=length,rank=rank,space=Space.Fourier)
        MA1['A','A'] = values2
        MA1['A','B'] = values1
        MA1['A','C'] = values1
        MA1['B','B'] = values3
        MA1['B','C'] = values3
        MA1['C','C'] = values3
        
        ntypes = len(types)
        PT = PairTable(types,'density')
        PT[['A'],['B','C']] = values1
        PT[['A'],['A']] = values2
        PT.setUnset(values3)
        MA2 = PT.exportToMatrixArray(space=Space.Fourier)
        
        np.testing.assert_array_almost_equal(MA1.data,MA2.data)
        self.assertEqual(MA1.space,MA2.space)
        
if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(PairTable_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python

from __future__ import division,print_function
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
from pyPRISM.core.PairTable import PairTable
import numpy as np

def spinodal_condition(PRISM,extrapolate=True):
    r'''Calculate the spinodal condition between pairs of components 

        
    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.

    extrapolate: bool, *optional*
        If *True*, only return the value extrapolated to :math:`k=0` rather than
        reporting the value at the lowest-k. Defaults to *True*.
    
    Returns
    -------
    lambda: pyPRISM.core.MatrixArray
        The full MatrixArray of structure factors
        

    **Mathematical Definition**

    .. math::

        \hat{\Lambda}_{\alpha,\beta}(k) =  1 & -\rho^{site}_{\alpha,\alpha} \hat{C}_{\alpha,\alpha}(k) \hat{\omega}_{\alpha,\alpha}(k) \\
        &  -2\rho^{site}_{\alpha,\beta} \hat{C}_{\alpha,\beta}(k) \hat{\omega}_{\alpha,\beta}(k) \\
        &  -\rho^{site}_{\beta,\beta} \hat{C}_{\beta,\beta} \hat{\omega}_{\beta,\beta}(k) \\
        &  +\rho^{site}_{\alpha,\beta} \rho^{site}_{\alpha,\beta} \hat{C}_{\alpha,\beta}(k)  \hat{C}_{\alpha,\beta}(k)  \hat{\omega}_{\alpha,\beta}(k) \hat{\omega}_{\alpha,\beta}(k) \\
        &  -\rho^{site}_{\alpha,\beta} \rho^{site}_{\alpha,\beta} \hat{C}_{\alpha,\alpha}(k) \hat{C}_{\beta,\beta}(k)  \hat{\omega}_{\alpha,\beta}(k) \hat{\omega}_{\alpha,\beta}(k) \\
        &  +\rho^{site}_{\alpha,\alpha} \rho^{site}_{\beta,\beta} \hat{C}_{\alpha,\alpha}(k) \hat{C}_{\beta,\beta}(k)  \hat{\omega}_{\alpha,\alpha}(k) \hat{\omega}_{\beta,\beta}(k) \\
        &  -\rho^{site}_{\alpha,\alpha} \rho^{site}_{\beta,\beta} \hat{C}_{\alpha,\beta}(k)  \hat{C}_{\alpha,\beta}(k)  \hat{\omega}_{\alpha,\alpha}(k) \hat{\omega}_{\beta,\beta}(k) \\

    **Variable Definitions**

        - :math:`\hat{\omega}_{\alpha,\beta}(k)`
            Intra-molecular correlation function between sites :math:`\alpha`
            and :math:`\beta` at a wavenumber :math:`k`

        - :math:`\hat{c}_{\alpha,\beta}(k)`
            Direct correlation function between sites :math:`\alpha` and
            :math:`\beta` at a wavenumber :math:`k`

        - :math:`\rho^{site}_{\alpha,\beta}` 
            Sitewise density for sites :math:`\alpha` and
            :math:`\beta`. See :class:`pyPRISM.core.Density` for details. 

    **Description**

        The spinodal condition (:math:`\hat{\Lambda}_{\alpha,\beta}(k)`) can be
        used to identify liquid-liquid macrophase separation between site types
        :math:`\alpha` and :math:`\beta` when
        :math:`\hat{\Lambda}_{\alpha,\beta}(k\rightarrow 0)=0`


    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.


    References
    ----------
    #. Schweizer, Curro, Integral equation theory of the structure and
       thermodynamics of polymer blends, J. Chem. Phys., 1989 91 (8) 5059 [`link
       <https://doi.org/10.1063/1.457598>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        spin = pyPRISM.calculate.spinodal_conditon(PRISM)

        spin_AB = spin['A','B']
    
    
    '''
    
    assert PRISM.sys.rank>1,'The spinodal calculation is only valid for multicomponent systems'
    
    if PRISM.directCorr.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.directCorr)

    if PRISM.omega.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.omega)
        
    
    lam = PairTable(name='spinodal_condition',types=PRISM.sys.types)
    for i,t1 in enumerate(PRISM.sys.types):
        for j,t2 in enumerate(PRISM.sys.types):
            if i<j:

                omega_AA  = PRISM.omega[t1,t1]
                omega_AB  = PRISM.omega[t1,t2]
                omega_BB  = PRISM.omega[t2,t2]

                C_AA = PRISM.directCorr[t1,t1]
                C_AB = PRISM.directCorr[t1,t2]
                C_BB = PRISM.directCorr[t2,t2]

                rho_AA = PRISM.sys.density.site[t1,t1]
                rho_AB = PRISM.sys.density.site[t1,t2]
                rho_BB = PRISM.sys.density.site[t2,t2]
               
                omega_AA *= 1.0/rho_AA
                omega_AB *= 1.0/rho_AB
                omega_BB *= 1.0/rho_BB
                
                curve  = +1
                curve += -1*C_AA * rho_AA * omega_AA
                curve += -2*C_AB * rho_AB * omega_AB
                curve += -1*C_BB * rho_BB * omega_BB
                curve += +C_AB*C_AB * rho_AB*rho_AB * omega_AB*omega_AB
                curve += -C_AA*C_BB * rho_AB*rho_AB * omega_AB*omega_AB
                curve += -C_AB*C_AB * rho_AA*rho_BB * omega_AA*omega_BB
                curve += +C_AA*C_BB * rho_AA*rho_BB * omega_AA*omega_BB

                fit = np.poly1d(np.polyfit(PRISM.sys.domain.k[:3],curve[:3],2))
                
                lam[t1,t2] = fit(0)
                  
        
    
    return lam



#!python
from __future__ import division,print_function
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
import warnings
import numpy as np

COMPONENT_WARNING = '''
This calculation was derived for a two component system. It is often the case that
these calculations can be generalized for pairs of sites withing multicomponent 
systems. We caution the user when interpreting the data from this calculation 
for more than two components. 
'''

def chi(PRISM,extrapolate=True):
    r'''Calculate the effective interaction parameter, :math:`\chi`

    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.

    extrapolate: bool, *optional*
        If *True*, only return the chi value extrapolated to :math:`k=0` rather
        than returning :math:`\chi(k)`
    
    Returns
    -------
    chi: pyPRISM.core.PairTable
        PairTable of all :math:`\chi(k)`  or :math:`\chi(k=0)` values
    

    **Mathematical Definition**

    .. math::
        
        \hat{\chi}_{\alpha,\beta}(k)  = \frac{0.5 \rho}{R^{+0.5} \phi_{\alpha} + R^{-0.5} \phi_{\beta}} (R^{-1} \hat{C}_{\alpha,\alpha}(k)
        + R \hat{C}_{\beta,\beta}(k) - 2 \hat{C}_{\alpha,\beta}(k))

    .. math::

        R = v_{\alpha}/v_{\beta}


    **Variable Definitions**

        - :math:`\hat{\chi}_{\alpha,\beta}(k)` 
            Wavenumber dependent effective interaction parameter between site
            types :math:`\alpha` and :math:`\beta`
    
        - :math:`\rho`
            Total system density from the :class:`pyPRISM.core.Density`
            instance stored in the system object (which is stored in the PRISM
            object)
    
        - :math:`\phi_{\alpha},\phi_{\beta}` 
            Volume fraction of site types :math:`\alpha` and :math:`\beta`. 
    
            .. math::
    
                \phi_{\alpha} = \frac{\rho_{\alpha}}{\rho_{\alpha} + \rho_{\beta}}

        - :math:`v_{\alpha},v_{\beta}` 
            Volume of site type :math:`\alpha` and :math:`\beta`


    **Description**

        :math:`\hat{\chi}_{\alpha,\beta}(k)` describes the overall effective
        interactions between site types :math:`\alpha` and :math:`\beta` as a
        single number. While there are many different definitions of
        :math:`\chi`, this is an effective version that takes into account both
        *entropic* and *enthalpic* interactions. In this way, this :math:`\chi`
        is similar to a second virial coefficient. In terms of value,
        :math:`\chi<0` indicates effective attraction and :math:`\chi>0`
        effective repulsion. 

        As most theories do not take into account the (potentially contentious)
        wavenumber dependence of :math:`\chi`, the zero-wavenumber extrapolation
        is often used when reporting PRISM-based :math:`\chi` values. For
        convenience, the full wavenumber dependent curve can be requested, but
        only the :math:`k=0` values are returned by default. 

    .. warning::

        The :math:`\chi` calculation is only valid for multicomponent systems
        i.e. systems with more than one defined type. This method will throw an
        exception if passed a 1-component PRISM object. 

    .. warning::

        This calculation is only rigorously defined in the two-component case.
        With that said, pyPRISM allows this method to be called for 
        multicomponent systems in order to calculate pairwise
        :math:`\chi` values. We urge caution when using this method for
        multicomponent systems as it is not clear if this approach is fully
        rigorous.

    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    
    References
    ----------
    Schweizer, Curro, Thermodynamics of Polymer Blends,
    J. Chem. Phys., 1989 91 (8) 5059, DOI: 10.1063/1.457598 [`link <http://dx.doi.org/10.1063/1.457598>`__]


    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        chi = pyPRISM.calculate.chi(PRISM)

        chi_AB = chi['A','B']
        chi_AA = chi['A','A'] #returns None because self-chi values are not defined

    '''
    
    assert PRISM.sys.rank>1,'The chi calculation is only valid for multicomponent systems'

    if PRISM.sys.rank!=2:
        warnings.warn(COMPONENT_WARNING)
    
    if PRISM.directCorr.space == Space.Real:
        PRISM.domain.MatrixArray_to_fourier(PRISM.directCorr)
        
    
    chi = PairTable(name='chi',types=PRISM.sys.types)
    chi0 = PairTable(name='chi0',types=PRISM.sys.types)
    for i,t1 in enumerate(PRISM.sys.types):
        for j,t2 in enumerate(PRISM.sys.types):
            if i<j:
                C_AA = PRISM.directCorr[t1,t1]
                C_AB = PRISM.directCorr[t1,t2]
                C_BB = PRISM.directCorr[t2,t2]

                v_A = 4.0/3.0 * np.pi * (PRISM.sys.diameter[t1]/2.0)**(3.0)
                v_B = 4.0/3.0 * np.pi * (PRISM.sys.diameter[t2]/2.0)**(3.0)

                rho_A = PRISM.sys.density[t1]
                rho_B = PRISM.sys.density[t2]

                phi_A = rho_A/(rho_A + rho_B)
                phi_B = rho_B/(rho_A + rho_B)

                R = v_A/v_B
                
                chi[t1,t2] = (R**(-0.5)*phi_A + R**(0.5)*phi_B)**(-1.0)*0.5*PRISM.sys.density.total*(R**(-1.0) * C_AA + R*C_BB - 2*C_AB)

                x = PRISM.sys.domain.k[:3]
                y = chi[t1,t2][:3]
                fit = np.poly1d(np.polyfit(x,y,2))
                chi0[t1,t2] = fit(0)

    if extrapolate:
        return chi0
    else:
        return chi



import warnings

try:
    import pint
except ImportError:
    warnings.warn('The Pint python library must be installed for the unit\
                   converter utility to be used. The UnitConverter class will not be\
                   available for use.')
else:
    from pyPRISM.util.UnitConverter import UnitConverter



#!python
from __future__ import division,print_function
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Domain import Domain
import unittest
import numpy as np

class Domain_TestCase(unittest.TestCase):
    def test_MatrixArray_loop(self):
        '''Can we transform an entire MatrixArray?'''
        length = 1024
        rank = 3
        dr = 0.1
        d = Domain(length=length,dr=dr)
        
        array1 = np.sin(np.arange(0,10*np.pi,0.01))[:length]
        array2 = 5*np.sin(np.arange(0,10*np.pi,0.01))[:length]
        array3 = np.cos(np.arange(0,10*np.pi,0.01))[:length]
        array4 = np.cos(np.arange(0,10*np.pi,0.01))[:length] + np.sin(np.arange(0,10*np.pi,0.01))[:length]
        
        MA = MatrixArray(length=length,rank=rank)
        MA['A','A'] = array1
        MA['B','B'] = array2
        MA['B','C'] = array3
        MA['C','C'] = array4
        
        d.MatrixArray_to_fourier(MA)
        np.testing.assert_array_almost_equal(MA['A','A'],d.to_fourier(array1))
        np.testing.assert_array_almost_equal(MA['B','B'],d.to_fourier(array2))
        np.testing.assert_array_almost_equal(MA['B','C'],d.to_fourier(array3))
        np.testing.assert_array_almost_equal(MA['C','B'],d.to_fourier(array3))
        np.testing.assert_array_almost_equal(MA['C','C'],d.to_fourier(array4))
        
        # This should recover the original values of the MatrixArray
        d.MatrixArray_to_real(MA)
        np.testing.assert_array_almost_equal(MA['A','A'],array1)
        np.testing.assert_array_almost_equal(MA['B','B'],array2)
        np.testing.assert_array_almost_equal(MA['B','C'],array3)
        np.testing.assert_array_almost_equal(MA['C','B'],array3)
        np.testing.assert_array_almost_equal(MA['C','C'],array4)
        
    def test_array_loop(self):
        '''Can we go to Fourier space and back again?'''
        d = Domain(length=1024,dr=0.1)
        
        real_space_data1 = np.sin(np.arange(0,10*np.pi,0.01))[:1024]
        
        fourier_space_data = d.to_fourier(real_space_data1)
        
        real_space_data2 = d.to_real(fourier_space_data)
        
        # If the DST coefficients are correct, we should get back (within numerical
        # precision) the same array we started with.
        np.testing.assert_array_almost_equal(real_space_data1,real_space_data2)

if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(Domain_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python 
from __future__ import division,print_function
import warnings
import numpy as np
from pyPRISM.core.PRISM import PRISM
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.ValueTable import ValueTable
from pyPRISM.core.Space import Space
from pyPRISM.core.Density import Density
from pyPRISM.core.Diameter import Diameter

from pyPRISM.closure.AtomicClosure import AtomicClosure
from pyPRISM.closure.MolecularClosure import MolecularClosure

class System:
    '''Primary class used to spawn PRISM calculations

    **Description**

        The system object contains tables that fully describe a system to be
        simulated. This includes the domain definition, all site densities,
        site diameters, interaction potentials, intra-molecular correlation
        functions (:math:`\hat{\omega}(k)`), and closures. This class also
        contains a convenience function for spawning a PRISM object. 
    
    
    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)

        sys.density['A'] = 0.1
        sys.density['B'] = 0.75

        sys.diameter[sys.types] = 1.0
        
        sys.closure[sys.types,sys.types] = pyPRISM.closure.PercusYevick()

        sys.potential[sys.types,sys.types] = pyPRISM.potential.HardSphere()
        
        sys.omega['A','A'] = pyPRISM.omega.SingleSite()
        sys.omega['A','B'] = pyPRISM.omega.InterMolecular()
        sys.omega['B','B'] = pyPRISM.omega.Gaussian(sigma=1.0,length=10000)
        
        PRISM = sys.createPRISM()

        PRISM.solve()

    '''
    def __init__(self,types,kT=1.0):
        r'''
        Arguments
        ---------
        types: list
            Lists of the site types that define the system

        kT: float
            Thermal temperature where :math:`k` is the Boltzmann constant and
            :math:`T` temperature. This is typicaly specified in reduced units
            where :math:`k_{B}=1.0`.


        Attributes
        ----------
        types: list
            List of site types
            
        rank: int
            Number of site types
        
        density: :class:`pyPRISM.core.Density`
            Container for all density values
            
        potential: :class:`pyPRISM.core.PairTable`
            Table of pair potentials between all site pairs in real space
            
        closure: :class:`pyPRISM.core.PairTable`
            Table of closures between all site pairs
            
        omega: :class:`pyPRISM.core.PairTable`
            Table of omega correlation functions in Fourier-space
        
        domain: :class:`pyPRISM.core.Domain`
            Domain object which specifies the Real and Fourier space 
            solution grid.
            
        kT: float
            Value of the thermal energy level. Used to vary temperature and
            scale the potential energy functions.

        diameter: :class:`pyPRISM.core.ValueTable`
            Site diameters. 

            .. warning::
            
                These diameters are currently only passed to the closures. They
                are not passed to potentials and it is up to the user to set
                sane \sigma values that match these diameters. 
        '''

        self.types = types
        self.rank  = len(types)
        self.kT = kT
        
        self.domain    = None
        self.diameter  = Diameter(types)
        self.density   = Density(types)
        self.potential = PairTable(types,'potential')
        self.closure   = PairTable(types,'closure')
        self.omega     = PairTable(types,'omega')

    def check(self):
        '''Is everything in the system specified?

        Raises
        ------
        ValueError if all values are not set
        
        '''
        for table in [self.density,self.potential,self.closure,self.omega,self.diameter]:
            table.check()
        
        if self.domain is None:
            raise ValueError(('System has no domain! '
                              'User must instatiate and assign a domain to the system!'))

        tol = 1e-6
        # ensure that the diameters are on the real-space grid
        for i,t,d in self.diameter.diameter:
            check = np.any(np.abs(self.domain.r - d)<tol)
            if not check:
                warn_text  = 'Diameter for site {} = {} is not a multiple '.format(t,d)
                warn_text += 'of domain grid spacing dr = {}. '.format(self.domain.dr)
                warn_text += 'Rounding will occur in closures, potentials, and omega!'
                warnings.warn(warn_text)

        # ensure that the sigmas are on the real-space grid
        for i,(t1,t2),s in self.diameter.sigma:
            check = np.any(np.abs(self.domain.r - s)<tol)
            if not check:
                warn_text  = 'Sigma for pair {}-{} = {} is not a multiple '.format(t1,t2,s)
                warn_text += 'of domain grid spacing dr = {}. '.format(self.domain.dr)
                warn_text += 'Rounding will occur in closures, potentials, and omega!'
                warnings.warn(warn_text)

    def createPRISM(self):
        '''Construct a PRISM object

        .. note::

            This method calls :func:`~pyPRISM.core.System.System.check` before creating the PRISM object.

        Returns
        -------
        PRISM: pyPRISM.core.PRISM
            Fully specified PRISM object
            
        '''
        self.check() #sanity check
        
        return PRISM(self)
    def solve(self,*args,**kwargs):
        '''Construct a PRISM object and attempt a numerical solution

        .. note::

            See :func:`~pyPRISM.core.PRISM.PRISM.solve` for arguments to this function

        .. note::

            This method calls :func:`~pyPRISM.core.System.System.check` before creating the PRISM object.
        
        Returns
        -------
        PRISM: pyPRISM.core.PRISM
            **Solved** PRISM object
            
        '''
        self.check() #sanity check

        p = PRISM(self)

        p.solve(*args,**kwargs)
        
        return p
        



#!python
'''
PRISM is often used in conjunction with molecular simulation techniques, such
as when using Self-Consistent PRISM. This module is intended to provide classes for
working with and analyzing molecular simulation trajectories.


See :ref:`scprism` for more information on the method.
'''
import warnings
from sys import platform as _platform


try:
    from pyPRISM.trajectory.Debyer import Debyer
except ImportError:
    warnings.warn('Cannot import Debyer: compiled Cython module not found.')
    '''
    See http://pyprism.readthedocs.io/en/latest/install/cython.plugins.html
    '''
else:
    if not (_platform == "linux" or _platform == "linux2"):
        warnings.warn('Parallelized Debyer is only supported on Linux. Using slower, serial execution.')
        



#!python
from __future__ import division,print_function
import pyPRISM
import numpy as np
import unittest

class CalcPRISM_TestCase(unittest.TestCase):
    def setup(self):
        '''Construct a simple but fully specified PRISM problem'''
        
        sys = pyPRISM.System(['A','B'])
        
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        
        sys.density['A'] = 0.1
        sys.density['B'] = 0.75

        sys.diameter[sys.types] = 1.0
        
        sys.closure.setUnset(pyPRISM.closure.PercusYevick())
        
        sys.potential.setUnset(pyPRISM.potential.HardSphere(sigma=1.0))
        
        sys.omega['A','A'] = pyPRISM.omega.SingleSite()
        sys.omega['A','B'] = pyPRISM.omega.NoIntra()
        sys.omega['B','B'] = pyPRISM.omega.Gaussian(sigma=1.0,length=10000)
        
        PRISM = sys.createPRISM()
        
        return PRISM

    def test_pair_correlation(self):
        '''Can we calculate pair_correlations?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.pair_correlation(PRISM)
        
    def test_structure_factor(self):
        '''Can we calculate structure factors?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.structure_factor(PRISM)
        
    def test_pmf(self):
        '''Can we calculate pmf's?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.pmf(PRISM)

    def test_chi(self):
        '''Can we calculate chi's?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.chi(PRISM)

    def test_second_virial(self):
        '''Can we calculate B2's?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.second_virial(PRISM)

    def test_solvation_potential(self):
        '''Can we calculate psi?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.solvation_potential(PRISM)

    def test_spinodal(self):
        '''Can we calculate spinodal conditions?'''
        PRISM = self.setup()
        PRISM.solve(options={'disp':False})
        result = pyPRISM.calculate.spinodal_condition(PRISM)
        
        
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(CalcPRISM_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
import numpy as np
class Closure:
    '''Baseclass for all closures 
    
    .. note::
    
        Currently, this class doesn't do anything besides group all of the
        closures under a single inheritance heirarchy. 
    '''
        



#!python
from __future__ import division,print_function
import unittest
from pyPRISM.potential.HardCoreLennardJones import HardCoreLennardJones
import numpy as np

class HardCoreLennardJones_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a HardCoreLennardJones potential?'''
        r = np.arange(0.75,3.5,0.05)
        epsilon = 0.25
        sigma   = 1.05
        high_value = 1e6
        
        U1 = epsilon*((sigma/r)**(12.0) - 2*(sigma/r)**(6.0))
        U1[r<=sigma] = high_value
        
        U2 = HardCoreLennardJones(epsilon,sigma,high_value=high_value).calculate(r)
        np.testing.assert_array_almost_equal(U1,U2)
        
        
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(HardCoreLennardJones_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Density import Density
import unittest
import numpy as np

class Density_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a density object? '''
        types = ['A','B','C']
        d = Density(types)

        d['A'] = 0.5
        d['B'] = 0.25
        d['C'] = 0.1

    def test_createDensityMatrices(self):
        '''Can we construct the two necesssary density matrices?'''

        d = Density(['A','B'])
        
        d['A'] = rhoA = 0.45
        d['B'] = rhoB = 0.35
        
        siteDensityMatrix1 = np.zeros((1,2,2))
        siteDensityMatrix1[:,0,0] = [rhoA]
        siteDensityMatrix1[:,0,1] = [rhoA + rhoB]
        siteDensityMatrix1[:,1,0] = [rhoA + rhoB]
        siteDensityMatrix1[:,1,1] = [rhoB]
        
        pairDensityMatrix1 = np.zeros((1,2,2))
        pairDensityMatrix1[:,0,0] = [rhoA * rhoA]
        pairDensityMatrix1[:,0,1] = [rhoA * rhoB]
        pairDensityMatrix1[:,1,0] = [rhoB * rhoA]
        pairDensityMatrix1[:,1,1] = [rhoB * rhoB]
        
        np.testing.assert_array_almost_equal(d.site.data,siteDensityMatrix1)
        np.testing.assert_array_almost_equal(d.pair.data,pairDensityMatrix1)

if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(Density_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
import unittest
import numpy as np
import pyPRISM

class PRISM_TestCase(unittest.TestCase):
    def setup(self):
        '''Construct a simple but fully specified PRISM problem'''
        
        sys = pyPRISM.System(['A','B'])
        
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        
        sys.density['A'] = 0.2
        sys.density['B'] = 0.6

        sys.diameter[sys.types] = 1.0
        
        sys.closure[sys.types,sys.types] = pyPRISM.closure.PercusYevick()
        
        sys.potential[sys.types,sys.types] = pyPRISM.potential.HardSphere(sigma=1.0)
        
        sys.omega['A','A'] = pyPRISM.omega.SingleSite()
        sys.omega['A','B'] = pyPRISM.omega.NoIntra()
        sys.omega['B','B'] = pyPRISM.omega.Gaussian(sigma=1.0,length=10000)
        
        PRISM = sys.createPRISM()
        
        return PRISM
    def test_solve(self):
        '''Can we solve the PRISM equations?'''
        PRISM = self.setup()
        result = PRISM.solve(options={'disp':False})
        self.assertIsNot(result,None)
        
        
        
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(PRISM_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function

from pyPRISM.core.Space import Space
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.IdentityMatrixArray import IdentityMatrixArray
from pyPRISM.closure.AtomicClosure import AtomicClosure
from pyPRISM.closure.MolecularClosure import MolecularClosure

from scipy.optimize import root

import numpy as np

from copy import deepcopy
import warnings

class PRISM:
    r'''Primary container for a storing a PRISM calculation
    
    Each pyPRISM.PRISM object serves as an encapsulation of a fully specified
    PRISM problem including all inputs needed for the calculation and the
    function to be numerically minimized. 
    
    Attributes
    ----------
    domain: pyPRISM.Domain
        The Domain object fully specifies the Real- and Fourier- space solution
        grids.
    
    directCorr: pyPRISM.MatrixArray
        The direct correlation function for all pairs of site types
        
    omega: pyPRISM.MatrixArray
        The intra-molecular correlation function for all pairs of site types
    
    closure: pyPRISM.core.PairTable of pyPRISM.closure.Closure
        Table of closure objects used to generate the direct correlation
        functions (directCorr)
        
    pairCorr: pyPRISM.MatrixArray
        The *inter*-molecular pair correlation functions for all pairs of
        site types. Also commonly refered to as the radial distribution functions.
    
    totalCorr: pyPRISM.MatrixArray
        The *inter*-molecular total correlation function is simply the pair
        correlation  function y-shifted by 1.0 i.e.  totalCorr = pairCorr - 1.0
        
    potential: pyPRISM.MatrixArray
        Interaction potentials for all pairs of sites
        
    GammaIn,GammaOut: pyPRISM.MatrixArray
        Primary inputs and outputs of the PRISM cost function. Gamma is defined as
        "totalCorr - directCorr" (in Fourier space) and results from a change
        of variables used to remove divergences in the closure relations. 
    
    OC,IOC,I,etc: pyPRISM.MatrixArray
        Various MatrixArrays used as intermediates in the PRISM functional.
        These arrays are pre-allocated and stored for efficiency. 
    
    x,y: float np.ndarray
        Current inputs and outputs of the cost function
    
    pairDensityMatrix: float np.ndarray
        Rank by rank array of pair densities between sites. See :class:`pyPRISM.core.Density`
            
    siteDensityMatrix: float np.ndarray
        Rank by rank array of site densities. See :class:`pyPRISM.core.Density`
    
    Methods
    -------
    cost:
        Primary cost function used to define the criteria of a "converged"
        PRISM solution. The numerical solver will be given this function 
        and will attempt to find the inputs (self.x) that make the outputs
        (self.y) as close to zero as possible. 
        
        
    '''
    def __init__(self,sys):
        self.sys = deepcopy(sys)

        # Need to set the potential for each closure object
        for (i,j),(t1,t2),U in self.sys.potential.iterpairs():
            if isinstance(self.sys.closure[t1,t2],AtomicClosure):
                #only set sigma if not set directly in potential
                if U.sigma is None:
                    U.sigma = self.sys.diameter[t1,t2]
                self.sys.closure[t1,t2].sigma = self.sys.diameter[t1,t2]
                self.sys.closure[t1,t2].potential = U.calculate(self.sys.domain.r) / self.sys.kT
            elif isinstance(self.sys.closure[t1,t2],MolecularClosure):
                raise NotImplementedError('Molecular closures are not fully implemented in this release.')
                #only set sigma if not set directly in potential
                if U.sigma is None:
                    U.sigma = self.sys.diameter[t1,t2]
                self.sys.closure[t1,t2].sigma = self.sys.diameter[t1,t2]
                self.sys.closure[t1,t2].potential = U.calculate_attractive(self.sys.domain.r) / self.sys.kT

        
        #cost function input and output
        self.x         = np.zeros(sys.rank*sys.rank*sys.domain.length)
        self.y         = np.zeros(sys.rank*sys.rank*sys.domain.length)

        # The omega objects must be converted to a MatrixArray of the actual correlation
        # function values rather than a table of OmegaObjects.
        applyFunc = lambda x: x.calculate(sys.domain.k)
        self.omega  = self.sys.omega.apply(applyFunc,inplace=False).exportToMatrixArray(space=Space.Fourier)
        self.omega *= sys.density.site #omega should always be scaled by site density 
        
        # Spaces are set based on when they are used in self.cost(...). In some cases,
        # this is redundant because these array's will be overwritten with copies and
        # then their space will be inferred from their parent MatrixArrays
        self.directCorr = MatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Real,types=sys.types)
        self.totalCorr  = MatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Fourier,types=sys.types)
        self.GammaIn    = MatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Real,types=sys.types)
        self.GammaOut   = MatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Real,types=sys.types)
        self.OC         = MatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Fourier,types=sys.types)
        self.I          = IdentityMatrixArray(length=sys.domain.length,rank=sys.rank,space=Space.Fourier,types=sys.types)
        
    def __repr__(self):
        return '<PRISM length:{} rank:{}>'.format(self.sys.domain.length,self.sys.rank)
        
    def cost(self,x):
        r'''Cost function 
        
        There are likely several cost functions that could be imagined using
        the PRISM equations. In this case we formulate a self-consistent 
        formulation where we expect the input of the PRISM equations to be
        identical to the output. 

        .. image:: ../../img/numerical_method.png
            :width: 300px
        
        The goal of the solve method is to numerically optimize the input (:math:`r \gamma_{in}`) 
        so that the output (:math:`r(\gamma_{in}-\gamma_{out})`) is minimized to zero.
        
        '''
        self.x = x #store input

        # The np.copy is important otherwise x saves state between calls to
        # this function.
        self.GammaIn.data = np.copy(x.reshape((-1,self.sys.rank,self.sys.rank)))
        self.GammaIn     /= self.sys.domain.long_r
        
        # directCorr is calculated directly in Real space but immediately 
        # inverted to Fourier space. We must reset this from the last call.
        self.directCorr.space = Space.Real 
        for (i,j),(t1,t2),closure in self.sys.closure.iterpairs():
            if isinstance(closure,AtomicClosure):
                self.directCorr[t1,t2] = closure.calculate(self.sys.domain.r,self.GammaIn[t1,t2])
            elif isinstance(closure,MolecularClosure):
                raise NotImplementedError('Molecular closures are untested and not fully implemented.')
                self.directCorr[t1,t2] = closure.calculate(self.GammaIn[t1,t2],self.omega[t1,t1],self.omega[t2,t2])
            else:
                raise ValueError('Closure type not recognized')
            
        self.sys.domain.MatrixArray_to_fourier(self.directCorr)
        
        self.OC = self.omega.dot(self.directCorr)
        self.IOC = self.I - self.OC
        self.IOC.invert(inplace=True)
        
        self.totalCorr  = self.IOC.dot(self.OC).dot(self.omega)
        self.totalCorr /= self.sys.density.pair
        
        self.GammaOut  = self.totalCorr - self.directCorr
        
        self.sys.domain.MatrixArray_to_real(self.GammaOut)
        
        self.y = self.sys.domain.long_r*(self.GammaOut.data - self.GammaIn.data)
        
        return self.y.reshape((-1,))
    def solve(self,guess=None,method='krylov',options=None):
        '''Attempt to numerically solve the PRISM equations
        
        Using the supplied inputs (in the constructor), we attempt to numerically
        solve the PRISM equations using the scheme laid out in :func:`cost`. If the 
        numerical solution process is successful, the attributes of this class
        will contain the solved values for a given input i.e. self.totalCorr will
        contain the numerically optimized (solved) total correlation functions.

        This function also does basic checks to ensure that the results are 
        physical. At this point, this consists of checking to make sure that
        the pair correlation functions are not negative. If this isn't true
        a warning is issued to the user. 
        
        Parameters
        ----------
        guess: np.ndarray, size (rank*rank*length)
            The initial guess of :math:`\gamma` to the numerical solution process.
            The numpy array should be of size rank x rank x length corresponding to 
            the a full flattened MatrixArray. If not specified, an initial guess
            of all zeros is used. 
            
        method: string
            Set the type of optimization scheme to use. The scipy documentation
            for `scipy.optimize.root
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html>`__
            details the possible values for this parameter. 
            

        options: dict
            Dictionary of options specific to the chosen solver method. The
            scipy documentation for `scipy.optimize.root
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html>`__
            details the possible values for this parameter. 
        
        '''
        if guess is None:
            guess = np.zeros(self.sys.rank*self.sys.rank*self.sys.domain.length)
            
        if options is None:
            options = {'disp':True}

        result = root(self.cost,guess,method=method,options=options)

        
        if self.totalCorr.space == Space.Fourier:
            self.sys.domain.MatrixArray_to_real(self.totalCorr)

        tol = 1e-5
        warnstr = 'Pair correlations are negative (value = {:3.2e}) for {}-{} pair!'
        for i,(t1,t2),H in self.totalCorr.itercurve():
            if np.any(H<-(1.0+tol)):
                val = np.min(H)
                warnings.warn(warnstr.format(val,t1,t2))
        
        return result
        
    



#!python
from __future__ import division,print_function
import numpy as np
class Omega(object):
    '''Baseclass for all *intra*-molecular correlation functions
    
    .. note::
    
        Currently, this class doesn't do anything besides group all of the
        *intra*-molecular correlation functions under a single inheritance
        heirarchy. This will change as needs arise.
    '''
        



#!python
from __future__ import division,print_function
from pyPRISM.closure.MolecularClosure import MolecularClosure
from pyPRISM.closure.PercusYevick import PercusYevick

from scipy.signal import fftconvolve
import numpy as np

class ReferenceMolecularPercusYevick(MolecularClosure):
    r'''Reference Molecular Percus Yevick (R-MPY) closure

    .. warning::

        This closure is not fully implemented. It should throw an error if a
        user attempts to use it.
    

    **Mathematial Definition**

        .. math:: 
            c_{\alpha,\beta}(r) = & \omega_{\alpha,\alpha}(r)*c^0_{\alpha,\beta}(r)*\omega_{\beta,\beta}(r) + \\
            & \omega_{\alpha,\alpha}(r)*[(\exp(-u_{\alpha,\beta}(r)) - 1.0) (1.0 + \gamma_{\alpha,\beta}(r))]*\omega_{\beta,\beta}(r) \\
            \gamma_{\alpha,\beta}(r) = & h_{\alpha,\beta}(r) - c_{\alpha,\beta}(r)

    
    **Variables Definitions**

        - :math:`A*B`
            Convolution integral between A and B

        - :math:`h_{\alpha,\beta}(r)` 
            Total correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`c_{\alpha,\beta}(r)`
            Direct correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.

        - :math:`c^0_{\alpha,\beta}(r)`
            Referene of Direct correlation function value at distance :math:`r`
            between sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`\omega_{\alpha,\beta}(r)`
            *Intra*-molecular correlation function value at distance :math:`r`
            between sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`u_{\alpha,\beta}(r)`
            Non-hardcore portion of Interaction potential value at distance
            :math:`r` between sites :math:`\alpha` and :math:`\beta`.
    
    **Description**

        TBA

    
    References
    ----------
        Schweizer, K.S.; Yethiraj, Arun; Polymer reference interaction site
        model theory: New molecular closure for phase separating fluids and
        alloys, The Journal of Chemical Physics 98, 9053 (1993); doi:
        10.1063/1.464465

    '''
    def __init__(self,C0):
        raise NotImplementedError('Molecular closures are untested and not fully implemented.')

        self._potential = None
        self.value = None
        self.PY = PercusYevick()
        self.C0 = C0
        
    def __repr__(self):
        return '<MolecularClosure: ReferenceMolecularPercusYevick>'


    @property
    def potential(self):
        return self._potential

    @potential.setter
    def potential(self,value):
        self._potential =  value
        self.PY.potential = value
    
    def calculate(self,gamma,omega1,omega2):
        r'''Calculate direct correlation function

        Arguments
        ---------
        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function

        omega1,omega2: np.ndarray
            array of :math:`\omega_{\alpha,\alpha}` or
            :math:`\omega_{\beta,\beta}` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'

        
        value = self.C0 + self.PY.calculate(gamma)

        self.value = fftconvolve(fftconvolve(omega1,value,mode='same'),omega2,mode='same')
        
        return self.value



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np

class FromArray(Omega):
    '''Read *intra*-molecular correlations from a list or array
    
    This class reads the omega from a Python list or Numpy array.
    
    Attributes
    ----------
    omega: np.ndarray
        Intra-molecular omega

    k: np.narray, *optional*
        Domain of the array data. If provided, this will be checked against the
        Fourier-space grid specified in the :class:`pyPRISM.core.Domain`. An
        exception will be raised is they do not match.

    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #set all omega(k) = 1 for type A
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        omega = np.ones(sys.domain.k.shape[0])
        sys.omega['A','A']  = pyPRISM.omega.FromArray(omega)
        x = sys.domain.k
        y = sys.omega['A','A'].calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()

    
    '''
    def __init__(self,omega,k=None):
        r'''Constructor
        
        Arguments
        ---------
        omega: list,np.ndarray
            Python list or Numpy array containing values of omega as
            a function of wavenumber :math:`k`.
            
        k: np.ndarray, *optional*
           Python list of Numpy array containing values of k. These must
           match the k values stored in the :class:`pyPRISM.core.Domain` or an
           exception will be raised. 
        '''
        self.value = np.array(omega)
        self.k = np.array(k)
        
    def __repr__(self):
        return '<Omega: FromArray>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        assert self.value.shape[0] == k.shape[0],'Size of array differs from domain!'
        if self.k is not None:
            assert self.k.shape[0] == k.shape[0],'File k-values differ from domain!'
            assert np.allclose(self.k,k),'File k-values differ from domain!'
        return self.value
        
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np

class NoIntra(Omega):
    '''Inter-molecule intra-molecular correlation function
    
    **Description**
        This is a convenience class for specifying the intra-molecular
        correlations between sites which are never in the same molecule.
        Because they have no *intra*-molecular correlation, this function
        returns zero at all wavenumber.
    
    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #set omega(k) for types A,B to have no intra-molecular
        #correlations (sites A and B are never on the same molecule)
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.omega['A','B']  = pyPRISM.omega.NoIntra()
        x = sys.domain.k
        y = sys.omega['A','B'].calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()
    
    '''
    def __repr__(self):
        return '<Omega: NoIntra>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        self.value = np.zeros_like(k)
        return self.value

        



#!python
from __future__ import division,print_function
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
from pyPRISM.calculate.pair_correlation import pair_correlation
import numpy as np

def pmf(PRISM):
    r'''Calculate the potentials of mean force

    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.
    
    Returns
    -------
    pmf: pyPRISM.core.MatrixArray
        The full MatrixArray of potentials of mean force

    
    **Mathematical Definition**

    .. math::

        w_{\alpha,\beta}(r) = -k_{B} T \ln(h_{\alpha,\beta}(r)+1.0)
    
    **Variable Definitions**

        - :math:`w_{\alpha,\beta}(r)`
            Potential of mean force between site types :math:`\alpha` and
            :math:`\beta` at a distance :math:`r`

        - :math:`g_{\alpha,\beta}(r)`
            Pair correlation function between site types :math:`\alpha` and
            :math:`\beta` at a distance :math:`r`

        - :math:`h_{\alpha,\beta}(r)`
            Total correlation function between site types :math:`\alpha` and
            :math:`\beta` at a distance :math:`r`

    **Description**

        A potential of mean force (PMF) between site types :math:`\alpha` and
        :math:`\beta`, :math:`w_{\alpha,\beta}` represents the the ensemble
        averaged free energy change needed to bring these two sites from
        infinite separation to a distance :math:`r`. It can also be thought of
        as a potential that would be needed to reproduce the underlying
        :math:`g_{\alpha,\beta}(r)`. 


    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        pmf = pyPRISM.calculate.pmf(PRISM)

        pmf_BB = pmf['B','B']
    
    
    '''
    rdf = pair_correlation(PRISM)
    
    #let's ignore any warnings about negative values in the log function
    with np.errstate(invalid='ignore'):
        rdf = -1.0 * PRISM.sys.kT * np.log(rdf.data)
    
    #length and rank will be inferred from data
    pmf = MatrixArray(data=rdf,
                      space=Space.Real,
                      length=None,
                      rank=None,
                      types=PRISM.sys.types)
    
    return pmf



#!python
from __future__ import division,print_function
from pyPRISM.closure.AtomicClosure import AtomicClosure
import numpy as np
import warnings
class MartynovSarkisov(AtomicClosure):
    r'''MartynovSarkisov closure 
    

    **Mathematial Definition**

        .. math::

            c_{\alpha,\beta}(r) = \left(\exp\left(\sqrt{\gamma_{\alpha,\beta}(r) - U_{\alpha,\beta}(r) - 0.5}\right) - 1.0 \right) - 1.0 -  \gamma_{\alpha,\beta}(r)

        .. math::
            
            \gamma_{\alpha,\beta}(r) =  h_{\alpha,\beta}(r) - c_{\alpha,\beta}(r)


    **Variables Definitions**

        - :math:`h_{\alpha,\beta}(r)` 
            Total correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`c_{\alpha,\beta}(r)`
            Direct correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`U_{\alpha,\beta}(r)`
            Interaction potential value at distance :math:`r` between sites
            :math:`\alpha` and :math:`\beta`.
    

    **Description**
        
        The Martynov-Sarkisov (MS) closure is described as a generalization of the
        HyperNettedChain closure. See the references below for derivation and
        usage examples.
        
        The change of variables is necessary in order to use potentials with
        hard cores in the computational setting. Written in the standard form,
        this closure diverges with divergent potentials, which makes it
        impossible to numerically solve. 
        
        The MS closure has been shown to be very accurate for hard-sphere
        spherical molecules and for high-density hard-core polymer systems.

    
    References
    ----------
    #. Martynov, G.A.; Sarkisov, G.N.; Mol. Phys. 49. 1495 (1983)
       [`link <https://doi.org/10.1080/00268978300102111>`__]

    #. Yethiraj, A.; Schweizer, K.S.; J. Chem. Phys. 97. 1455 (1992)
       [`link <https://doi.org/10.1063/1.464465>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        sys.closure['A','A'] = pyPRISM.closure.PercusYevick()
        sys.closure['A','B'] = pyPRISM.closure.PercusYevick()
        sys.closure['B','B'] = pyPRISM.closure.MartynovSarkisov()

        # ** finish populating system object **

        PRISM = sys.createPRISM()

        PRISM.solve()
    
    '''
    def __init__(self,apply_hard_core=False):
        '''Contstructor

        Parameters
        ----------
        apply_hard_core: bool
            If *True*, the total correlation function will be assumed to be -1
            inside the core (:math:`r_{i,j}<(d_i + d_j)/2.0`) and the closure
            will not be applied in this region.
        '''
        self.potential = None
        self.value = None
        self.apply_hard_core = apply_hard_core

        if apply_hard_core == False:
            warnings.warn(
                    '''The MartynovSarkisov closure does not work for divergent potentials
                    when the hard core condition is not manually applied. This
                    will likely result in a cryptic crash of the simulation if
                    attempted. Using MartynovSarkisov(apply_hard_core=True) will avoid this
                    warning. This warning should be ignored if hard-core
                    interactions are not being used.'''
                    )
        
    def __repr__(self):
        return '<AtomicClosure: MartynovSarkisov>'
    
    def calculate(self,r,gamma):
        '''Calculate direct correlation function based on supplied :math:`\gamma`

        Arguments
        ---------
        r: np.ndarray
            array of real-space values associated with :math:`\gamma`

        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'
        
        
        if self.apply_hard_core:
            assert self.sigma is not None, 'If apply_hard_core=True, sigma parameter must be set!'

            # apply hard core condition 
            self.value = -1 - gamma

            # calculate closure outside hard core
            mask = r>self.sigma
            self.value[mask] = np.exp(np.sqrt(gamma[mask] - self.potential[mask] + 0.5) - 1.0) - 1.0 - gamma[mask]
        else:
            self.value = np.exp(np.sqrt(gamma - self.potential + 0.5) - 1.0) - 1.0 - gamma

        
        return self.value
        
        
class MS(MartynovSarkisov):
    '''Alias of MartynovSarkisov'''
    pass



#!python
from __future__ import division,print_function
from pyPRISM.omega.GaussianRing import GaussianRing
import unittest
import numpy as np

class GaussianRing_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a GaussianRing Omega?'''
        k = np.arange(0.75,3.5,0.05)
        sigma = 1.0
        length = N = 100
        
        
        O1 = np.zeros_like(k)
        ss = sigma * sigma
        kk = k*k
        for i in range(length):
            j = 0
            O1 += np.exp(-ss*kk*abs(i-j)*(length-abs(i-j))/(6.0*length))
        
        G = GaussianRing(sigma,length)
        O2 = G.calculate(k)
        np.testing.assert_array_almost_equal(O1,O2)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.NoIntra import NoIntra

class InterMolecular(NoIntra):
    '''alias of NoIntra intra-molecular correlation function '''
    def __repr__(self):
        return '<Omega: InterMolecular>'
    
        
        



#  tests for pyprism-1.0.3-py310h2873277_1004 (this is a generated file);
print('===== testing package: pyprism-1.0.3-py310h2873277_1004 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import pyPRISM
pyPRISM.test()
#  --- run_test.py (end) ---

print('===== pyprism-1.0.3-py310h2873277_1004 OK =====');
print("import: 'pyPRISM'")
import pyPRISM

print("import: 'pyPRISM.core'")
import pyPRISM.core

print("import: 'pyPRISM.calculate'")
import pyPRISM.calculate

print("import: 'pyPRISM.closure'")
import pyPRISM.closure

print("import: 'pyPRISM.omega'")
import pyPRISM.omega

print("import: 'pyPRISM.util'")
import pyPRISM.util

print("import: 'pyPRISM.potential'")
import pyPRISM.potential

print("import: 'pyPRISM.trajectory'")
import pyPRISM.trajectory




#!python
from __future__ import division,print_function
import numpy as np
class Potential(object):
    '''Baseclass for all intermolecular-pairwise potentials
    
    ..note:: 
    
        Currently, this class doesn't do anything besides group all of the
        potentials under a single inheritance heirarchy. 
    
    '''
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.SingleSite import SingleSite
import unittest
import numpy as np

class SingleSite_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a SingleSite Omega?'''
        k = np.arange(0.75,3.5,0.05)
        O1 = np.ones_like(k)
        
        O2 = SingleSite().calculate(k)
        np.testing.assert_array_almost_equal(O1,O2)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.potential.Potential import Potential
import numpy as np
class HardSphere(Potential):
    r'''Simple hard sphere potential
    

    **Mathematical Definition**
    
    .. math::
    
        U_{\alpha,\beta}(r>\sigma_{\alpha,\beta}) = 0.0
    
    .. math::
    
        U_{\alpha,\beta}(r\leq\sigma_{\alpha,\beta}) = C^{high}
    
    
    **Variable Definitions**
    
    :math:`\sigma_{\alpha,\beta}`
        Length scale of interaction between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`r`
        Distance between sites. 

    :math:`C^{high}`
        High value used to approximate an infinite potential due to overlap
    
   
    **Description**

        This potential models the simiple hard-sphere fluid, in which sites
        have hard-core repulsion and no interactions outside their contact
        distance.
    
    Example
    -------
    .. code-block:: python

        import pyPRISM
	
        #Define a PRISM system and set the A-B interaction potential
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.potential['A','B'] = pyPRISM.potential.HardSphere(sigma=1.0,high_value=10**6)

    .. warning::

        If sigma is specified such that it does not fall on the solution grid
        of the :class:`~pyPRISM.core.Domain.Domain` object specified in
        :class:`~pyPRISM.core.System.System`, then the sigma will effectively
        be rounded. A warning should be emitted during the construction of a
        :class:`~pyPRISM.core.PRISM.PRISM` object if this occurs.
    
    '''
    def __init__(self,sigma=None,high_value=1e6):
        r''' Constructor
        
        Arguments
        ---------
        sigma: float, *optional*
            Contact distance. If not specified, sigma will be calculated from 
            the diameters specified in the :class:`~pyPRISM.core.System.System`
            object.
            
        high_value: float, *optional*
            High value used to approximate an infinite potential due to overlap
        '''
        self.sigma = sigma
        self.high_value = high_value
        self.funk  = lambda r,sigma: np.where(r>sigma,0.0,high_value)
    def __repr__(self):
        return '<Potential: HardSphere>'
    
    def calculate(self,r):
        r'''Calculate value of potential

        Attributes
        ----------
        r: float np.ndarray
            Array of pair distances at which to calculate potential values
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'

        magnitude = self.funk(r,self.sigma)
        return magnitude
        



#!python
from __future__ import division,print_function
import unittest
from pyPRISM.potential.HardSphere import HardSphere
import numpy as np

class HardSphere_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a HardSphere potential?'''
        r = np.arange(0.75,3.5,0.05)
        sigma   = 1.05
        high_value = 1e5
        
        U1 = np.ones_like(r)*high_value
        U1[r>sigma] = 0.0
        
        U2 = HardSphere(sigma,high_value).calculate(r)
        np.testing.assert_array_almost_equal(U1,U2)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.potential.Potential import Potential
import numpy as np

class LennardJones(Potential):
    r'''12-6 Lennard-Jones potential
    

    **Mathematical Definition**
    
    .. math::
    
        U_{\alpha,\beta}(r) = 4\epsilon_{\alpha,\beta}\left[\left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{12.0} - \left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{6.0}\right]

    .. math::
    
        U_{\alpha,\beta}^{shift}(r) = U_{\alpha,\beta}(r) - U_{\alpha,\beta}(r_{cut})
    
    
    **Variable Definitions**
    
    :math:`\epsilon_{\alpha,\beta}`
        Strength of attraction between sites :math:`\alpha` and :math:`\beta`.

    :math:`\sigma_{\alpha,\beta}`
        Length scale of interaction between sites :math:`\alpha` and
        :math:`\beta`.

    :math:`r`
        Distance between sites. 
    
    :math:`r_{cut}`
        Cutoff distance between sites. 
   
    **Description**

        The classic 12-6 LJ potential. To facilitate direct comparison with
        molecular simulation, the simulation may be cut and shifted to zero at
        a specified cutoff distance by setting the rcut and shift parameters.
        The full (non-truncated) LJ potential is accessed using
        :math:`r_{cut}` =  *None* and :math:`shift` = *False*.
    
    
    Example
    -------
    .. code-block:: python

        import pyPRISM

        #Define a PRISM system and set the A-B interaction potential
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.potential['A','B'] = pyPRISM.potential.LennardJones(epsilon=1.0,sigma=1.0,rcut=2.5,shift=True)

    .. warning::

        If sigma is specified such that it does not fall on the solution grid
        of the :class:`~pyPRISM.core.Domain.Domain` object specified in
        :class:`~pyPRISM.core.System.System`, then the sigma will effectively
        be rounded. A warning should be emitted during the construction of a
        :class:`~pyPRISM.core.PRISM.PRISM` object if this occurs.

    
    '''
    def __init__(self,epsilon,sigma=None,rcut=None,shift=False):
        r''' Constructor
        
        Arguments
        ---------
        epsilon: float
            Depth of attractive well
            
        sigma: float, *optional*
            Contact distance. If not specified, sigma will be calculated from 
            the diameters specified in the :class:`~pyPRISM.core.System.System`
            object.
            
        rcut: float, *optional*
            Cutoff distance for potential. Useful for comparing directly to results
            from simulations where cutoffs are necessary. 
    
        shift: bool,*optional*
            If :math:`r_{cut}` is specified, shift the potential by its value
            at the cutoff. If :math:`r_{cut}` is not specified, this parameter
            is ignored.
    
        '''
        self.epsilon = epsilon
        self.sigma = sigma
        self.rcut  = rcut
        self.shift = shift
        self.funk  = lambda r,s: 4 * epsilon * ((s/r)**(12.0) - (s/r)**(6.0))
        
    def __repr__(self):
        return '<Potential: LennardJones>'
        
    def calculate(self,r):
        r'''Calculate value of potential

        Attributes
        ----------
        r: float np.ndarray
            Array of pair distances at which to calculate potential values
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'

        magnitude = self.funk(r,self.sigma)
        
        if self.rcut is not None:
            if self.shift:
                magnitude -= self.funk(self.rcut,self.sigma)
            magnitude[r>self.rcut] = 0.0
                
        return magnitude

    def calculate_attractive(self,r):
        r'''Calculate the attractive tail of the Lennard Jones potential. Returns zero at :math:`r<\sigma`
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'

        magnitude = np.zeros_like(r)
        mask = r>self.sigma
        magnitude[mask] = self.calculate(r,self.sigma)[mask]
        return magnitude
        



import unittest
from pyPRISM.core.ValueTable import ValueTable
import numpy as np

class ValueTable_TestCase(unittest.TestCase):
    def test_get_set(self):
        '''Can we set and get values from the table?'''
        VT = ValueTable(['A','B','C'],'density')
        VT['A'] = 0.4
        VT[['B','C']] = 0.25
        
        self.assertEqual(VT['A'],0.4)
        self.assertEqual(VT['B'],0.25)
        self.assertEqual(VT['C'],0.25)
        
    def test_check(self):
        '''Can we check to make sure the table is filled?'''
        VT = ValueTable(['A','B','C'],'density')
        VT[['A','B']] = 0.4
        self.assertRaises(ValueError,VT.check)
        
        
    def test_setUnset(self):
        '''Can wet set all of the unset table values?'''
        VT = ValueTable(['A','B','C'],'density')
        VT['A'] = 0.4
        VT.setUnset(0.25)
        self.assertEqual(VT['A'],0.4)
        self.assertEqual(VT['B'],0.25)
        self.assertEqual(VT['C'],0.25)
        
        
    def test_iter(self):
        '''Can we iterate over the table?'''
        types = ['A','B','C']
        ntypes = len(types)
        VT = ValueTable(types,'density')
        
        alphaTypes = ['A','B','C']
        
        counter = 0
        for i,t,v in VT:
            if t in alphaTypes:
                alphaTypes.remove(t)
            counter+=1
            
        #did we visit all expected pairs?
        self.assertEqual(len(alphaTypes),0)
        
        #sanity check, did we visit the correct number of pairs?
        self.assertEqual(counter,ntypes)
        
        
        
        
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.FromFile import FromFile
import unittest
import numpy as np
import os

class FromFile_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a FromFile Omega?'''
        k = np.arange(0.75,3.5,0.05)
        fname = 'FromFileTestCase.dat'
        
        np.savetxt(fname,k)
        
        FF = FromFile(fname).calculate(k)
        
        os.remove(fname)
    def test_error_check(self):
        '''Does FromFile correctly check the domain?'''
        fname = 'FromFileTestCase.dat'
        
        k = np.arange(0.75,3.5,0.05)
        dummy = np.array([k,k]).T
        
        np.savetxt(fname,dummy)
        
        k = np.arange(0.75,3.5,0.01) #change the asked for domain...
        with self.assertRaises(AssertionError):
            FF = FromFile(fname).calculate(k)
        
        os.remove(fname)
        
        



#!python
r'''
While the PRISM equation specifies the base PRISM formalism, we need additional
equations called closures to numerically solve the PRISM equations for
:math:`\hat{H}(k)` and :math:`\hat{C}(k)`. Closures provide a mathematical
relation between the direct correlation function :math:`c(r)`, the pairwise
interaction potential :math:`u(r)`, and, often, the total correlation function
:math:`h(r)`. Since the closures include :math:`u(r)`, it is through these
closures that the chemical details of the system are specified. 

'''
from pyPRISM.closure.Closure import Closure
from pyPRISM.closure.AtomicClosure import AtomicClosure
from pyPRISM.closure.MolecularClosure import MolecularClosure

from pyPRISM.closure.PercusYevick import PercusYevick
from pyPRISM.closure.PercusYevick import PY

from pyPRISM.closure.HyperNettedChain import HyperNettedChain
from pyPRISM.closure.HyperNettedChain import HNC

from pyPRISM.closure.MeanSphericalApproximation import MeanSphericalApproximation
from pyPRISM.closure.MeanSphericalApproximation import MSA

from pyPRISM.closure.MartynovSarkisov import MartynovSarkisov
from pyPRISM.closure.MartynovSarkisov import MS

from pyPRISM.closure.ReferenceMolecularPercusYevick import ReferenceMolecularPercusYevick




#!python
from __future__ import division,print_function
import pyPRISM
import unittest
import numpy as np

class System_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a system?'''
        types = ['A','B','C']
        sys = pyPRISM.System(types)
    def test_check(self):
        '''Can we validate a system object?'''
        types = ['A','B','C']
        sys = pyPRISM.System(types)
        self.assertRaises(ValueError,sys.check)
        
        # These values make no sense but their sufficient
        # for this 'check'
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.density[sys.types] = 1.0
        sys.diameter[sys.types] = 1.0
        sys.potential.setUnset(1.0)
        sys.closure.setUnset(1.0)
        sys.omega.setUnset(1.0)
        
        #test should fail if this raises
        try:
            sys.check()
        except ValueError:
            self.fail('Check failed when it should have passed...')
    def test_createPRISM(self):
        '''Can we construct a simple but fully specified PRISM problem?'''
        
        sys = pyPRISM.System(['A','B'])
        
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        
        sys.density['A'] = 0.45
        sys.density['B'] = 0.35

        sys.diameter[sys.types] = 1.0
        
        sys.closure.setUnset(pyPRISM.closure.PercusYevick())
        
        sys.potential.setUnset(pyPRISM.potential.HardSphere(sigma=1.0))
        
        sys.omega['A','A'] = pyPRISM.omega.SingleSite()
        sys.omega['A','B'] = pyPRISM.omega.NoIntra()
        sys.omega['B','B'] = pyPRISM.omega.Gaussian(sigma=1.0,length=100)
        
        PRISM = sys.createPRISM()
        
        self.assertIsInstance(PRISM,pyPRISM.core.PRISM.PRISM)
        
        
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(System_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.closure import PercusYevick
import unittest
import numpy as np

class PercusYevick_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create and use a PercusYevick closure?'''
        r = np.arange(0.75,3.5,0.05)
        U = np.zeros_like(r)
        U[r<1.0] = 1e6
        
        gamma = np.zeros_like(r)
        C1 = (np.exp(-U) - 1.0)*(1.0 + gamma)
        
        
        PY = PercusYevick(apply_hard_core=False)
        PY.potential = U
        PY.sigma = 1.0
        C2 = PY.calculate(r,gamma)
        np.testing.assert_array_almost_equal(C1,C2)
    def test_create_hard_core(self):
        '''Can we create and use a hard-core PercusYevick closure?'''
        r = np.arange(0.75,3.5,0.05)
        U = np.zeros_like(r)
        U[r<1.0] = 1e6
        sigma = 1.0
        
        gamma = np.zeros_like(r)
        C1 = -1 - gamma
        mask = r>sigma
        C1[mask] = (np.exp(-U[mask]) - 1.0)*(1.0 + gamma[mask])


        PY = PercusYevick(apply_hard_core=True)
        PY.potential = U
        PY.sigma = sigma
        C2 = PY.calculate(r,gamma)
        np.testing.assert_array_almost_equal(C1,C2)
        
        
if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(PercusYevick_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
import unittest
from pyPRISM.potential.LennardJones import LennardJones
import numpy as np

class LennardJones_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a LennardJones potential?'''
        r = np.arange(0.75,3.5,0.05)
        epsilon = 0.25
        sigma   = 1.05
        rcut    = 2.5*sigma
        shift   = True
        
        UShift = 4*epsilon*((sigma/rcut)**(12.0) - (sigma/rcut)**(6.0))
        U1 = 4*epsilon*((sigma/r)**(12.0) - (sigma/r)**(6.0)) - UShift
        U1[r>rcut] = 0
        
        U2 = LennardJones(epsilon,sigma,rcut=rcut,shift=shift).calculate(r)
        np.testing.assert_array_almost_equal(U1,U2)
        
        



from pyPRISM.omega.NoIntra import NoIntra
import unittest
import numpy as np

class NoIntra_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a NoIntra Omega?'''
        k = np.arange(0.75,3.5,0.05)
        O1 = np.zeros_like(k)
        
        O2 = NoIntra().calculate(k)
        np.testing.assert_array_almost_equal(O1,O2)
        
        



class Table:
    '''Baseclass used to define tables of parameters
    
    .. note::

        This class should not be used/instatiated directly. It is only intended to be
        inherited.
    
    '''
    def listify(self,values):
        '''Helper fuction that converts any input into a list of inputs.
        
        The purpose of this function is to help with iterating over types,
        and to handle the case of a single :class:`str` type being passed. 
        '''
        if isinstance(values,str):
            values = [values]
        else:
            try:
                iter(values)
            except TypeError:
                values = [values]
            else:
                values = list(values)
        return values
        



#!python
from __future__ import division,print_function
from pyPRISM.closure import MeanSphericalApproximation
import unittest
import numpy as np

class MeanSphericalApproximation_TestCase(unittest.TestCase):
    def test_create_hard_core(self):
        '''Can we create and use a hard-core MeanSphericalApproximation closure?'''
        r = np.arange(0.75,3.5,0.05)
        U = np.zeros_like(r)
        U[r<1.0] = 1e6
        sigma = 1.0
        
        gamma = np.zeros_like(r)
        C1 = -1 - gamma
        mask = r>sigma
        C1[mask] = -U[mask]


        MSA = MeanSphericalApproximation(apply_hard_core=True)
        MSA.potential = U
        MSA.sigma = sigma
        C2 = MSA.calculate(r,gamma)
        np.testing.assert_array_almost_equal(C1,C2)
        
        
if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(MeanSphericalApproximation_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np

class FromFile(Omega):
    '''Read *intra*-molecular correlations from file
    
    This class reads a one or two column file using Numpy's
    loadtxt function. A one-column file is expected to only
    contain intra-molecular correlation data, while a two-column
    file contains the k values of the data in the first column
    as well. If the k-values are provided, an extra check to make
    sure that the file data matches the same k-space grid of the
    domain. 
    
    Attributes
    ----------
    fileName: str
        full path + name to column file
        
    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #set type A omega(k) from a file
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        fileName = './test_example_filename.txt'
        sys.omega['A','A']  = pyPRISM.omega.FromFile(fileName)
        x = sys.domain.k
        y = sys.omega['A','A'].calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()

    
    '''
    def __init__(self,fileName):
        r'''Constructor
        
        Arguments
        ---------
        fileName: str
            path to textfile containing values of omega as
	    a function of wavenumber, k.
            
        '''
        self.fileName = fileName
        
    def __repr__(self):
        return '<Omega: FromFile>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        fileData = np.loadtxt(self.fileName)
        
        if len(fileData.shape)>=2:
            assert fileData.shape[0] == k.shape[0],'Domain size of file differs from supplied domain!'
            assert np.allclose(fileData[:,0],k),'Domain of file differs from supplied domain!'
            self.value = fileData[:,1]
        else:
            self.value = fileData
        
        return self.value
        
        



from pyPRISM.core.Table import Table
import numpy as np

class ValueTable(Table):
    '''Container for data that is keyed by types


    **Description**

        The goal of this class is to provide a simple inteface for setting and
        storing parameters that are accessed and identified by types. This is
        typically site properties, e.g. density, site diameter. By default the
        value for all types is set to `None` and therefore can be checked to see
        if the table has been fully specified. 

        Setter/getter methods have been create to set groups of types
        simultaneously. This allows for the rapid construction of datasets
        where many of the parameters are repeTated. 

        Note that, unlike the :class:`pyPRISM.core.MatrixArray`, this
        container is not meant to be used for mathematics. The benefit of this
        is that, for each type, it can contain any arbitrary number, string, or
        Python object. 

        See the example below and the `pyPRISM Internals` section of the
        :ref:`tutorial` for more information.

    Example
    -------
    .. code-block:: python

        import pyPRISM

        VT = pyPRISM.ValueTable(['A','B','C','D','E'],name='density')

        # set the value for type A to be 0.25
        VT['A'] = 0.25

        # set the value for types B & C to be 0.35
        VT[ ['B','C'] ] = 0.35

        # set all other values to be 0.1
        VT.setUnset(0.1)

        for i,t,v in VT:
            print('{}) {} for type {} is {}'.format(i,VT.name,t,v))


        # The above loop prints the following:
        #   0) density for type A is 0.25
        #   1) density for type B is 0.35
        #   2) density for type C is 0.35
        #   3) density for type D is 0.1
        #   4) density for type E is 0.1
    
    
    '''
    def __init__(self,types,name):
        r'''Constructor
        
        Arguments
        ---------
        types: list
            Lists of the types that will be used to key the ValueTable. The
            length of this list should be equal to the rank of the PRISM
            problem to be solved i.e.  len(types) == number of sites in system.
            
        name: string
            The name of the ValueTable. Currently, this is simply used as a
            convencience for identifying the table internally. 
        '''

        self.types = types
        self.name = name
        self.values = {t:None for t in types}
    
    def __repr__(self):
        return '<ValueTable: {}>'.format(self.name)
        
    def __iter__(self):
        '''Data iterator

        This magic-method allows for ValueTables to be iterated over via
        `for x in y` constructs like

        .. code-block:: python

            for index,type,value in ValueTable: 
                print(index,type,value)

        Yields
        ------
        index: int
            index of value

        type: 
            type of value

        value: 
            stored value at this type
        '''
        for i,t in enumerate(self.types):
            yield i,t,self.values[t]
            
    def __getitem__(self,index):
        t = index
        return self.values[t]
    
    def __setitem__(self,index,value):
        types1 = index
        for t in self.listify(types1):
            self.values[t] = value
            
    def check(self):
        '''Is everything in the table set?

        Raises
        ------
        *ValueError* if all values are not set
        
        '''
        for i,t,val in self:
            if val is None:
                raise ValueError('ValueTable {} is not fully specified!'.format(self.name))
            
            
    def setUnset(self,value):
        '''Set all values that have not been specified to a value

        Arguments
        ---------
        value: 
            Any valid python object (number, list, array, etc) can be passed in
            as a value for all unset fields. 
        
        '''
        for i,t,v in self:
            if v is None:
                self[t] = value



#!python
from __future__ import division,print_function
from pyPRISM.core.ValueTable import ValueTable
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
import numpy as np

class Diameter(object):
    r'''Container for site diameters and sigmas

    **Mathematical Definition**

    .. math::
        
        \sigma_{\alpha,\beta} = (d_{\alpha} + d_{\beta})/2.0
        

    **Variable Definitions**

        :math:`d_{\alpha}`
            diameter of site-type :math:`\alpha`

        :math:`sigma_{\alpha,\beta}`
            contact distanct between site-type :math:`\alpha` and :math:`\beta`

    **Description**

        A simple container class for defining the site-type diameters. 
    
    

    Example
    -------
    .. code-block:: python

        import pyPRISM

        d = pyPRISM.Diameter(['A','B','C'])

        d['A'] = 1.0
        d['B'] = 1.5
        d['C'] = 5.0

        d.diameter['A']  #site diameter diam_A = 1.0
        d.sigma['A','B'] #site contact distance sig_AB = 1.25
        d.sigma['A','C'] #site contact distance sig_AB = 3.0
    
    '''
    def __init__(self,types):
        r'''Constructor 

        Arguments
        ---------
        types: list 
            List of types of sites
        
        Attributes
        ----------
        diameter: :class:`pyPRISM.core.ValueTable`
            Table of site site diameter values

        volume: :class:`pyPRISM.core.ValueTable`
            Table of site site volumes values

        sigma: :class:`pyPRISM.core.PairTable`
            Site-site contact distance 
        '''
        self.types = types 

        self.diameter = ValueTable(types=types,name='diameter')
        self.volume = ValueTable(types=types,name='volume')
        self.sigma = PairTable(types=types,name='sigma')

    def check(self):
        '''Are all diameter set?

        Raises
        ------
        *ValueError* if diameters are not all set. 
        
        '''
        self.diameter.check()

    def __repr__(self):
        return '<Diameter>'

    def __getitem__(self,key):
        key = self.diameter.listify(key)
        if len(key) == 1:
            return self.diameter[key[0]]
        elif len(key) == 2:
            return self.sigma[key[0],key[1]]
        else:
            ValueError('Too many types passed to diameter!')

    def __setitem__(self,types1,value):
        for t1 in self.diameter.listify(types1):
            d1 = value
            self.diameter[t1] = d1

            self.volume[t1] = (4.0/3.0) * np.pi * (d1/2.0)**(3.0)

            self.total = 0.
            for t2 in self.types:
                # If d2 isn't set yet, we can't set sigma
                d2 = self.diameter[t2]
                if d2 is None:
                    continue

                self.sigma[t1,t2] = (d1 + d2)/2.0




#!python
from __future__ import division,print_function
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Diameter import Diameter
import unittest
import numpy as np

class Diameter_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a diameter object? '''
        types = ['A','B','C']
        d = Diameter(types)

        d['A'] = 0.5
        d['B'] = 0.25
        d['C'] = 0.1

        np.testing.assert_almost_equal(d['A'],0.5)
        np.testing.assert_almost_equal(d['B'],0.25)
        np.testing.assert_almost_equal(d['C'],0.1)


    def test_sigma(self):
        '''Can we create the sigma matrix? '''
        types = ['A','B','C']
        d = Diameter(types)

        d['A'] = 0.5
        d['B'] = 0.25
        d['C'] = 0.1

        SigmaMatrix1 = np.zeros((3,3))
        for (i,j),t,v in d.sigma:
            SigmaMatrix1[i,j] = v

        SigmaMatrix2 = np.zeros((3,3))
        SigmaMatrix2[0,0] = (d['A']+d['A'])/2.0
        SigmaMatrix2[0,1] = (d['A']+d['B'])/2.0
        SigmaMatrix2[0,2] = (d['A']+d['C'])/2.0
        SigmaMatrix2[1,0] = (d['B']+d['A'])/2.0
        SigmaMatrix2[1,1] = (d['B']+d['B'])/2.0
        SigmaMatrix2[1,2] = (d['B']+d['C'])/2.0
        SigmaMatrix2[2,0] = (d['C']+d['A'])/2.0
        SigmaMatrix2[2,1] = (d['C']+d['B'])/2.0
        SigmaMatrix2[2,2] = (d['C']+d['C'])/2.0

        np.testing.assert_array_almost_equal(SigmaMatrix1,SigmaMatrix2)



if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(Diameter_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np

class SingleSite(Omega):
    '''Single-site intra-molecular correlation function
    
    This class is useful for dealing with single bead molecules
    such as solvents or large spherical particles, it sets the value
    of the intra-molecular correlation function to 1 for all wavenumbers. 
    
    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np

        #set omega(k) for type A to 1 (single spherical site per molecule)
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
	sys.omega['A','A']  = pyPRISM.omega.SingleSite()
        x = sys.domain.k
        y = sys.omega['A','A'].calculate(x)

        #plot using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()
    
    '''
    def __repr__(self):
        return '<Omega: SingleSite>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        self.value = np.ones_like(k)
        return self.value
        
        



#!python
from pyPRISM.core.Space import Space
import string
from itertools import product
import numpy as np

class MatrixArray(object):
    '''A container for creating and interacting with arrays of matrices
    
    **Description**

        The primary data structure of MatrixArray is simply a 3D Numpy array 
        with the first dimension accessing each individual matrix in the array
        and the last two dimenions corresponding to the vertical and horizontal 
        index of each matrix element.
        
        The terminology *Curve* is used to refer to the set of values from
        all matrices in the array at a given matrix index pair. In Numpy slicing 
        parlance::
        
            curve_11 = numpy_array[:,1,1]
            curve_12 = numpy_array[:,1,2]

        Access to the MatrixArray is either by supplied types or numerical indices.
        If types are not supplied, captial letters starting from 'A' are used. 

        See the example below and the `pyPRISM Internals` section of the
        :ref:`tutorial` for more information.

    Example
    -------
    .. code-block:: python

        mArray = MatrixArray(length=1024,rank=2,types=['polymer','solvent'])
        
        mArray['polymer','solvent'] == mArray['solvent','polymer'] == mArray.get(0,1)

    
    '''
    
    SpaceError = "Attempting MatrixArray math in non-matching spaces"
    
    def __init__(self,length,rank,data=None,space=Space.Real,types=None):
        '''Constructor

        Arguments
        ----------
        length: int
            Number of matrices in array. For PRISM theory, this corresponds to
            the number of grid points in real- and Fourier-space i.e.
            Domain.size.

        rank: int
            Number of rows/cols of each (square) matrix. For PRISM theory, this
            also equal to the number of site types.
            
        data: np.ndarray, size (length,rank,rank)
            Interface for specifying the MatrixArray data directly. If not
            given, all values in all matrices will be set to zero. 
            
        space: pyPRISM.core.Space.Space
            Enumerated value tracking whether the array represents real or
            Fourier spaced data. As we will be transferring arrays to and from
            these spaces, it's important for safety that we track this.

        types: list, *optional*
            List of semantic types that are be used to reference data. These
            types will be output by the itercurve method as well. If not
            supplied, uppercase letters will be used.

        '''
                    
        if data is None:
            self.data = np.zeros((length,rank,rank))
            self.rank = rank
            self.length = length
        else:
            assert len(data.shape)==3,'Data passed to MatrixArray must be 3-D'
            assert data.shape[1]==data.shape[2],'Last two dimensions of MatrixArray data must be same size'
            self.data = data
            self.rank = data.shape[1]
            self.length = data.shape[0]
        
        if types is None:
            self.types = list(string.ascii_uppercase[:self.rank])
        else:
            assert len(types)==self.rank
            self.types = types

        self.typeMap = {t:i for i,t in enumerate(self.types)}
        self.space = space

    def __repr__(self):
        return '<MatrixArray rank:{:d} length:{:d}>'.format(self.rank,self.length)
    
    def get_copy(self):
        '''Return an independent copy of this MatrixArray'''
        return MatrixArray(length=self.length,rank=self.rank,data=np.copy(self.data),space=self.space,types=self.types)
    
    def itercurve(self):
        '''Iterate over the curves in this MatrixArray

        Yields
        ------
        (i,j): 2-tuple of integers
            numerical index to the underlying data numpy array

        (t1,t2): 2-tuple of string types
            string index to the underlying data numpy array

        curve: np.ndarray, size (self.length)
            1-D array representing a curve within the MatrixArray
        '''
        for i,j in product(range(self.rank),range(self.rank)):
            if i<=j: #upper triangle condition
                type1 = self.types[i]
                type2 = self.types[j]
                yield (i,j),(type1,type2),self.data[:,i,j]
            
    def __setitem__(self,key,val):
        '''Curve setter 

        Arguments
        ---------
        key: tuple of types
            Type pair used to identify curve pair

        val: np.ndarray
            Values of curve


        Assumes all matrices are symmetric and enforces symmetry by
        setting the off-diagonal elements to be equal.

        '''
        type1,type2 = key 

        try:
            index1 = self.typeMap[type1]
        except KeyError:
            raise ValueError('This MatrixArray has types: {}. You requested type: \'{}\''.format(self.types,type1))

        try:
            index2 = self.typeMap[type2]
        except KeyError:
            raise ValueError('This MatrixArray has types: {}. You requested type: \'{}\''.format(self.types,type2))

        self.data[:,index1,index2] = val
        if not (index1 == index2):
            self.data[:,index2,index1] = val
        
    def __getitem__(self,key):
        '''Curve getter

        Arguments
        ---------
        key: tuple of types
            Type pair used to identify curve pair

        val: np.ndarray
            Values of curve
        '''

        type1,type2 = key 

        try:
            index1 = self.typeMap[type1]
        except KeyError:
            raise ValueError('This MatrixArray has types: {}. You requested type: \'{}\''.format(self.types,type1))

        try:
            index2 = self.typeMap[type2]
        except KeyError:
            raise ValueError('This MatrixArray has types: {}. You requested type: \'{}\''.format(self.types,type2))

        return self.data[:,index1,index2]
    
    def get(self,index1,index2):
        '''Curve getter via indices

        This method should be slightly more efficient than the standard
        __getitem__. 
        '''
        assert index1<self.rank,'Supplied index out of range'
        assert index2<self.rank,'Supplied index out of range'
        return self.data[:,index1,index2]

    def getMatrix(self,matrix_index):
        '''Matrix getter via indices'''
        return self.data[matrix_index,:,:]

    def setMatrix(self,matrix_index,value):
        '''Matrix setter via indices'''
        self.data[matrix_index,:,:] = value
    

    def __truediv__(self,other):
        '''Scalar or elementwise division'''
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            data = self.data / other.data
        else:
            data = self.data / other
        return MatrixArray(length=self.length,rank=self.rank,data=data,space=self.space,types=self.types)

    def __div__(self,other):
        return self.__truediv__(other)

    def __itruediv__(self,other):
        '''Scalar or elementwise division'''
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            self.data /= other.data
        else:
            self.data /= other
        return self

    def __idiv__(self,other):
        return self.__itruediv__(other)
    
    
    def __mul__(self,other):
        '''Scalar or elementwise multiplication'''
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            data = self.data * other.data
        else:
            data = self.data * other
        return MatrixArray(length=self.length,rank=self.rank,data=data,space=self.space,types=self.types)
    
    def __imul__(self,other):
        '''Scalar or elementwise multiplication'''
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            self.data *= other.data
        else:
            self.data *= other
        return self
    
    def __add__(self,other):
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            data = self.data + other.data
        else:
            data = self.data + other
        return MatrixArray(length=self.length,rank=self.rank,data=data,space=self.space,types=self.types)
    
    def __iadd__(self,other):
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            self.data += other.data
        else:
            self.data += other
        return self
            
    def __sub__(self,other):
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            data = self.data - other.data
        else:
            data = self.data - other
        return MatrixArray(length=self.length,rank=self.rank,data=data,space=self.space,types=self.types)
    
    def __isub__(self,other):
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
            self.data -= other.data
        else:
            self.data -= other
        return self
            
    def invert(self,inplace=False):
        '''Perform matrix inversion on all matrices in the MatrixArray
        
        Parameters
        ----------
        inplace: bool
            If False, a new MatrixArray is returned, otherwise just
            update the internal data.
        '''
        if inplace:
            data = self.data
        else:
            data = np.copy(self.data)
            
        data = np.linalg.inv(self.data)
            
        if inplace:
            self.data = data
            return self
        else:
            return MatrixArray(rank=self.rank,length=self.length,data=data,space=self.space,types=self.types)
        
    def dot(self,other,inplace=False):
        ''' Matrix multiplication for each matrix in two MatrixArrays
        
        Parameters
        ----------
        other: object, MatrixArray
            Must be an object of MatrixArray type of the same length
            and dimension
            
        inplace: bool
            If False, a new MatrixArray is returned, otherwise just
            update the internal data.
        
        '''
        if isinstance(other,MatrixArray):
            assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
        if inplace:
            self.data = np.einsum('lij,ljk->lik', self.data, other.data)
            return self
        else:
            data = np.einsum('lij,ljk->lik', self.data, other.data)
            return MatrixArray(length=self.length,rank=self.rank,data=data,space=self.space,types=self.types)
        
    def __matmul__(self,other):
        assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
        return self.dot(other,inplace=False)
        
    def __imatmul__(self,other):
        assert (self.space == other.space) or (Space.NonSpatial in (self.space,other.space)),MatrixArray.SpaceError
        return self.dot(other,inplace=True)
        
        



#!python

try:
    from enum import Enum
except ImportError:
    # Python 2.x
    parent = object
else:
    # Python 3.x
    parent = Enum


class Space(parent):
    ''' An enumeration to track which space an object is in

    **Description**
        MatrixArrays can represent data in Real- or Fourier- space and they can be
        transformed in-place between these spaces. This class is meant to help
        track this state by creating a standard numerical reference for each state
        that can be checked. This allows classes like
        :class:`pyPRISM.core.MatrixArray` to do error checking when doing math
        between arrays. This enumeration also defines a 'wildcard' state so that we
        can still do math with non-spatial data. 

    Example
    -------
    .. code-block:: python
        
        import pyPRISM

        A = pyPRISM.MatrixArray(length=1000,rank=3,space=pyPRISM.Space.Real)
        B = pyPRISM.MatrixArray(length=1000,rank=3,space=pyPRISM.Space.Real)
        C = pyPRISM.MatrixArray(length=1000,rank=3,space=pyPRISM.Space.Fourier)

        A.space == B.Space # returns True
        A.space == C.Space # returns False

        A.dot(C) #raises exception 

    .. note::

        The enumerated states of the Space Enum are listed below. The actual
        values of these states are not-important beyond being unique from one
        another.

    '''
    Real       = 1
    Fourier    = 2
    NonSpatial = 3
